import gzip
import multiprocessing
import shutil
import os
import sys
from .shared import *
from contextlib import contextmanager
from pudb.remote import set_trace

primer3_col_names = [
    'PRIMER_PAIR_0_PRODUCT_SIZE',
    'PRIMER_PAIR_0_PENALTY',
    'PRIMER_LEFT_0_SEQUENCE', 'PRIMER_RIGHT_0_SEQUENCE',
    'PRIMER_LEFT_0_PENALTY', 'PRIMER_RIGHT_0_PENALTY',
    'PRIMER_LEFT_0_TM', 'PRIMER_RIGHT_0_TM',
    'PRIMER_LEFT_0_GC_PERCENT', 'PRIMER_RIGHT_0_GC_PERCENT',
    'PRIMER_LEFT_0_SELF_ANY_TH', 'PRIMER_RIGHT_0_SELF_ANY_TH',
    'PRIMER_LEFT_0_SELF_END_TH', 'PRIMER_RIGHT_0_SELF_END_TH',
    'PRIMER_LEFT_0_HAIRPIN_TH', 'PRIMER_RIGHT_0_HAIRPIN_TH',
    'PRIMER_LEFT_0_END_STABILITY', 'PRIMER_RIGHT_0_END_STABILITY',
    'PRIMER_PAIR_0_COMPL_ANY_TH', 'PRIMER_PAIR_0_COMPL_END_TH',
]
primer3_col_key = {n: n.replace("PRIMER_", "").replace("_0", "").lower() for n in primer3_col_names}


def _render_csv_header(stream, primer3=False, out_sep=','):
    names = ["left_seq", "diag_seq", "right_seq"]
    if primer3:
        names.extend([primer3_col_key[n] for n in primer3_col_names])
    #set_trace(term_size=(80, 60))
    print(*names, sep=out_sep, file=stream)


def _format_p3_output(p3_out):
    """Reformat data for best primer pair for CSV output"""
    return {primer3_col_key[n]: p3_out[n] for n in primer3_col_names}


@contextmanager
def stream_writer(file_path=None, default_stream=sys.stdout, mode="a"):
    file_handle = default_stream if file_path is None else open(file_path, mode)
    yield file_handle
    if file_path is not None and file_handle is not None:
        file_handle.close()


def safe_print(alignments, csv_rows, out_align, out_csv, counter, lock, find_primers):
    """ Function to safely print alignments to the terminal in parallel """
    # Acquire lock
    with lock:
        # Iterate through alignments and print, increment counter
        if out_align is not None and out_csv is not None:
            for align, row in zip(alignments, csv_rows):
                print(align, file=out_align, flush=True)
                print(row, file=out_csv, flush=True)
                counter.value += 1
        elif out_csv is not None:
            for row in csv_rows:
                print(row, file=out_csv, flush=True)
                counter.value += 1
        elif out_align is not None:
            for align in alignments:
                print(align, file=out_align, flush=True)
                counter.value += 1


def render_output_part(kmerfile, out_align, out_csv, counter, lock,
                       start=None, end=None, print_block=100, ingroup=None, find_primers=False):
    """ Helper function to render the contents of a part of the kmer file """
    # Get alignment stream
    alignments = alignmentStream(kmerfile, start, end, ingroup)
    alignments_to_print = []
    row_to_print = []
    print_block_counter = 0
    with stream_writer(out_align, None) as out_align_stream,\
         stream_writer(out_csv, sys.stdout) as out_csv_stream:
        for alignment in alignments:
            # Run primer3 (NOTE: ideally this would run on its own step in krisp.main, not here)
            if find_primers:
                if not alignment.find_primers():
                    continue
            # Add to alignments list
            if out_align_stream is not None:
                alignments_to_print.append(alignment.render_alignment())
            # Add csv rows to list
            if out_csv_stream is not None:
                row_to_print.append(alignment.render_csv())
            #set_trace(term_size=(80, 60))
            # Print
            print_block_counter += 1
            if print_block_counter >= print_block:
                # Print alignments
                safe_print(alignments_to_print, row_to_print, out_align_stream, out_csv_stream, counter, lock, find_primers)
                alignments_to_print = []
                row_to_print = []
                print_block_counter = 0
        if print_block_counter > 0:
            safe_print(alignments_to_print, row_to_print, out_align_stream, out_csv_stream, counter, lock, find_primers)


def render_output(kmerfile, out_align=None, out_csv=sys.stdout, cores=1, print_block=10, ingroup=None, find_primers=False):
    """ Function to print alignments from a file in parallel

    Parameters
    ----------

    kmerfile : str
        The input kmerfile to render

    out_align : str or stream
        The path to the alignment output file or stream to write to (e.g. sys.stdout)

    out_csv : str or stream
        The path to the csv output file or stream to write to (e.g. sys.stdout)

    cores : int
        The number of processors to use

    print_block : int
        The number of results to buffer in memory before attempting to acquire the lock and write the results.

    ingroup : list of str
        The names of the in-group samples

    Returns
    -------
    int
        The number of alignments written

    """
    # Print CSV header if needed
    with stream_writer(out_csv, sys.stdout, mode="w") as out_csv_stream:
        _render_csv_header(out_csv_stream, primer3=find_primers)

    # Print alignment header if needed
    if isinstance(out_align, str) and os.path.isfile(out_align):
        os.remove(out_align)

    # Get the start and end iteration points for kmerfile
    fptr_start_end = list(splitFilePtrs(kmerfile, num_sections=cores))

    # Create a counter and lock
    counter = multiprocessing.Value('i', 0)
    lock = multiprocessing.Lock()

    # Create a list of job args
    job_args = []
    for start, end in fptr_start_end:
        job_args.append((kmerfile, out_align, out_csv, counter, lock, start, end, print_block, ingroup, find_primers))

    # Start a group of processes
    processes = []
    for job in job_args:
        p = multiprocessing.Process(target=render_output_part,
                                    args=job)
        p.start()
        processes.append(p)

    # Call join on processes
    for p in processes:
        p.join()
    return counter.value



# def writeAlignments(kmerfile, output, pid, start=None, end=None, ingroup=None):
#     """ Helper function to write alignments from kmer file """
#     # Get alignment stream
#     alignments = alignmentStream(kmerfile, start, end, ingroup)
#
#     # Open for writing and write output
#     with gzip.open(output, 'wt') as fout:
#         found = 0
#         for i, alignment in enumerate(alignments):
#             # Print allignment
#             fout.write(f"> process_{pid}_{i}\n")
#             fout.write(f"{str(alignment)}\n")
#             found += 1
#         fout.flush()
#     return found
#
#
# def writeAlignmentsParallel(kmerfile, output, parallel, workdir=None, ingroup=None):
#     """ Function to write alignments from a kmer file in parallel
#
#     Parameters
#     ----------
#
#     kmerfile : str
#         The input kmerfile to convert to an alignment file
#
#     output : str
#         The output alignment file
#
#     parallel : int
#         The number of processors to use
#
#     workdir : str (default None)
#         The working directory to place temporary files
#
#     Returns
#     -------
#     int
#         The number of alignments written
#
#     """
#
#     # Get the start and end iteration points for kmerfile
#     fptr_start_end = list(splitFilePtrs(kmerfile, parallel))
#
#     # Create a list of temporary output files
#     tmp_files = [tmpFile(workdir) for i in range(parallel)]
#
#     # Create a list of job args
#     job_args = []
#     for i in range(parallel):
#         start, end = fptr_start_end[i]
#         pid = i
#         tmp = tmp_files[i]
#         job_args.append((kmerfile, tmp, pid, start, end, ingroup))
#
#     # Run processes in pool
#     with multiprocessing.Pool(parallel) as pool:
#         counts = pool.starmap(writeAlignments, job_args)
#
#     # Merge tmpfiles into output
#     if output.endswith('.gz'):
#         with open(output, "wb") as fout:
#             for tmp in tmp_files:
#                 with open(tmp, "rb") as fin:
#                     shutil.copyfileobj(fin, fout)
#     else:
#         with open(output, "w") as fout:
#             for tmp in tmp_files:
#                 with gzip.open(tmp, "rt") as fin:
#                     shutil.copyfileobj(fin, fout)
#     return sum(counts)



from diagvar.find_diag_var import *
from diagvar.find_diag_var import _parse_group_data, _check_variant, GroupedVariant
import pysam
from itertools import islice
import unittest


class TestGroupedVariant(unittest.TestCase):
    
    def setUp(self):
        pysam.set_verbosity(0)
        self.vcf = pysam.VariantFile('test_data/unfilt_allscafs_n666.vcf.gz')
        self.groups = _parse_group_data('test_data/test_metadata.tsv')
        self.groups = {k: v for k, v in self.groups.items() if k in ['NA1','NA2', 'EU1', 'EU2']}
        self.variant1 = GroupedVariant(next(self.vcf), self.groups)
        self.variant2 = GroupedVariant(next(self.vcf), self.groups)
        self.variant3 = GroupedVariant(next(self.vcf), self.groups)
        self.diag_var = GroupedVariant(next(islice(self.vcf, 1020, 1021)), self.groups)
        self.vcf_subset = islice(self.vcf, 10, 15)

    def test_init(self):
        self.assertEqual(self.diag_var.sample_counts,
                         {'NA1': 222, 'EU1': 202, 'EU2': 0, 'NA2': 18})
        self.assertEqual(self.diag_var.allele_counts,
                         {'NA1': {'T': 222}, 'EU1': {'C': 202}, 'EU2': {}, 'NA2': {'C': 18}})
        self.assertEqual(self.diag_var.conserved,
                         {'NA1': 'T', 'EU1': 'C', 'EU2': None, 'NA2': 'C'})
        self.assertEqual(self.diag_var.diagnostic,
                         {'NA1': None, 'EU1': None, 'EU2': None, 'NA2': None})

    def test_from_vcf(self):
        gen = GroupedVariant.from_vcf(self.vcf_subset, groups=self.groups)
        self.assertEqual(next(gen).allele_counts,
                         {'NA1': {'T': 236}, 'EU1': {'C': 207}, 'EU2': {}, 'NA2': {'C': 22}})


if __name__ == '__main__':
    unittest.main()



import itertools
import subprocess
import multiprocessing
import shutil
import sys
import os
import time
from colorama import Fore, Back, Style
from .shared import *


def bisect(filename, key):
    """ Find the largest pointer value s.t. ptr.readline() < key

    Parameters
    ----------
    filename : str
        Name of file to bisect
    key : str
        Find the largest pointer value s.t. ptr.readline() < key

    Returns
    -------
    int
         Byte index for start of line

    """
    # Open filename for reading
    with open(filename, "r") as fptr:
        # Get size of file in bytes
        file_size = fileSize(fptr)

        # Check for Null input
        if key == "":
            return file_size, ""

        # bisect
        last_success = 0
        bracket = (0, file_size)
        prev_bracket = None
        while (bracket != prev_bracket):
            # Get starting values
            start, end = bracket

            # Get midpoint
            midpoint = (start + end) // 2

            # Move ptr to next full line
            fptr.seek(midpoint)
            junk = fptr.readline()
            midpoint = fptr.tell()

            # Update start, end
            line = fptr.readline()
            line = line.split(',')[0]
            if line < key:
                start = midpoint
                last_success = start
            else:
                end = midpoint

            # Update bracket
            prev_bracket = bracket
            bracket = (start, end)
        fptr.seek(last_success)
        return last_success, fptr.readline()


def mergeKmerFiles(filename0, filename1, output,
                   f0_start=None, f0_end=None, f1_start=None):
    """ Merge files in an output file

    Parameters
    ----------
    filename0 : str
        Name of first kmer file to merge
    filenam1 : str
        Name of second kmer file to merge
    output : str
        Name of output file to write
    f0_start : int, optional
        Start byte of first kmer file
    f0_end : int, optional
        End byte of first kmer file
    f1_start : int, optional
        Start byte of second kmer file

    Returns
    -------
    None
        Output file is written, None is returned
    """
    # generate alignment streams
    stream0 = alignmentStream(filename0, f0_start, f0_end)
    stream1 = alignmentStream(filename1, f1_start)

    # Take intersection of alignments
    alignments = intersectSortedStreams(stream0, stream1)

    # Write alignments back to file
    writeAlignmentStream(alignments, output)


def mergeKmerSingleCore(filename0, filename1, output):
    """ Merge two kmer files to output file in parallel

    Parameters
    ----------
    filename0 : str
        Name of first kmer file to merge
    filename1 : str
        Name of second kmer file to merge
    output : str
        Name of output file to write
    workdir : str, optional
        Name of work directory to write temporary files

    Returns
    -------
    None
        Output file is written, None is returned
    """
    # Now merge these files together using cat
    with open(output, "w") as fout:
        # Merge
        mergeKmerFiles(filename0, filename1, output)
        # Return merged filename
        return output


def mergeKmerParallel(filename0, filename1, output,
                      parallel, workdir=None):
    """ Merge two kmer files to output file in parallel

    Parameters
    ----------
    filename0 : str
        Name of first kmer file to merge
    filename1 : str
        Name of second kmer file to merge
    output : str
        Name of output file to write
    workdir : str, optional
        Name of work directory to write temporary files

    Returns
    -------
    None
        Output file is written, None is returned
    """
    # Start by making a list of temporary files
    tmp_files = [tmpFile(workdir) for i in range(parallel)]

    # Get the start and end iteration points for filename0
    f0_start_end = list(splitFilePtrs(filename0, parallel))

    # Find the bisection starting point for filename1
    f1_start = []
    with open(filename0) as fptr:
        for start, end in f0_start_end:
            # Get line at start
            fptr.seek(start)
            line = fptr.readline().split(',')[0]

            # Bisect on line and add pos to other_start
            pos, line2 = bisect(filename1, line)
            f1_start.append(pos)

    # Create a list of job args
    job_args = []
    for x, y, z in zip(tmp_files, f0_start_end, f1_start):
        job_args.append((filename0, filename1, x, *y, z))

    # Run jobs in parallel pool
    with multiprocessing.Pool(parallel) as pool:
        pool.starmap(mergeKmerFiles, job_args)

    # Now merge these files together using cat
    with open(output, "w") as fout:
        # Merge tmp_files into fptr.name
        for tmp in tmp_files:
            with open(tmp) as fin:
                fin.flush()
                shutil.copyfileobj(fin, fout)
                fout.flush()
        # Return merged filename
        return output


def mergeKmerJob(in_queue, out_queue, verbose=True):
    """ Process job for merging kmer files

    Parameters
    ----------
    in_queue : Multiprocessing Queue
        Input queue which holds jobs to run
    out_queue : Multiprocessing Queue
        Output queue to put finished jobs

    """
    # Get first job and iterate until None is encountered
    job_args = in_queue.get()
    while job_args is not None:
        # Split into args
        filename0, filename1, output, parallel, workdir = job_args

        # Print verbosity
        start_t = None
        if verbose:
            message = (f"Merging: {filename0} + {filename1} -> {output} using"
                       f" {parallel} cores")
            print(message, file=sys.stderr)
            start_t = time.time()

        # Merge files
        # Note that mergeKmerSingleCore is a temporary replacement for mergeKmerParallel which does not produce consistent results currently. See issue 17 on github for more details.
        #mergeKmerParallel(filename0, filename1, output, parallel, workdir)
        mergeKmerSingleCore(filename0, filename1, output)

        # Print verbosity
        if verbose:
            end_t = time.time()
            end_message = (f"=> Merged {filename0} + {filename1} -> {output}"
                           f" in {prettyTime(end_t-start_t)}")
            print(Fore.GREEN + end_message + Style.RESET_ALL, file=sys.stderr)

        # Add to output queue and get next job
        out_queue.put(output)
        job_args = in_queue.get()


def mergeFiles(files, output, parallel=1, workdir=None, verbose=True):
    """ Iteratively merges files until only one file is left

    Parameters
    ----------
    files : [str, str, ...]
        List of kmer filenames to merge
    output : str
        Ouptut file to write results
    parallel : int, optional
        Number of cores to use on this job
    wordir : str, optional
        Work directory to write temporary files
    verbose : bool, optional
        Print progress to stderr

    Returns
    -------
    None
        Output file is written
    """
    # Iterate through files and merge until only one file remains
    tmp_files = []
    counter = 0
    while len(files) > 1:
        # Get number of jobs to run and the number of cores per job
        jobs = len(files) // 2
        job_cores = []
        if parallel < jobs:
            job_cores = [1] * jobs
        else:
            for i in range(jobs):
                np = (parallel // jobs) + (i < parallel % jobs)
                job_cores.append(np)

        # Initialize processors with input and output queue
        processes = []
        job_queue = multiprocessing.Queue()
        fin_queue = multiprocessing.Queue()
        for i in range(jobs):
            p = multiprocessing.Process(target=mergeKmerJob,
                                        args=(job_queue, fin_queue, verbose))
            p.start()
            processes.append(p)

        # Get pairs of files to merge
        file_pairs = []
        while len(files) > 1:
            file_pairs.append((files.pop(), files.pop()))

        # Add jobs to queue
        for f0_f1, np in zip(file_pairs, job_cores):
            # Create temporary output file
            tmp_file = tmpFile(workdir)

            # Add job to queue
            f0, f1 = f0_f1
            job_queue.put((f0, f1, tmp_file, np, workdir))

        # now wait for jobs to finish and add to output list
        results = []
        for p in processes:
            result = fin_queue.get()
            results.append(result)
            tmp_files.append(result)

        # Send kill signal to processes and then join
        for p in processes:
            job_queue.put(None)

        # Call join on processes
        for p in processes:
            p.join()

        # Set new files to merge
        files = results + files

    # move result to output
    shutil.move(files[0], output)



import logging
import os.path
import sys
import argparse
import gzip
import primer3
import pysam
import re
import itertools
import multiprocessing as mp
import queue

from collections import deque, defaultdict
from Bio import SeqIO
from Bio.Data import IUPACData
from contextlib import contextmanager
from logging.handlers import QueueHandler
from statistics import mean
from nltk.metrics import distance

from .find_diag_var import GroupedVariant, _parse_group_data
from .print_align import render_variant, Annotation

# Constants
SNP_DELIM = ('<', '>')
PRIMER_DELIM = ('(', ')')
CRRNA_DELIM = ('{', '}')
HETERO_DELIM = '/'
UNKNOWN_CHAR = '?'
iupac_key = {tuple((x for x in sorted(v))): k for k, v in
             IUPACData.ambiguous_dna_values.items()}
iupac_key[(UNKNOWN_CHAR,)] = 'N'
# primer3_col_names = [
#     'PRIMER_PAIR_0_PRODUCT_SIZE',
#     'PRIMER_PAIR_0_PENALTY',
#     'PRIMER_LEFT_0_SEQUENCE', 'PRIMER_RIGHT_0_SEQUENCE', 'PRIMER_INTERNAL_0_SEQUENCE',
#     'PRIMER_LEFT_0_PENALTY', 'PRIMER_RIGHT_0_PENALTY', 'PRIMER_INTERNAL_0_PENALTY',
#     'PRIMER_LEFT_0_TM', 'PRIMER_RIGHT_0_TM', 'PRIMER_INTERNAL_0_TM',
#     'PRIMER_LEFT_0_GC_PERCENT', 'PRIMER_RIGHT_0_GC_PERCENT', 'PRIMER_INTERNAL_0_GC_PERCENT',
#     'PRIMER_LEFT_0_SELF_ANY_TH', 'PRIMER_RIGHT_0_SELF_ANY_TH', 'PRIMER_INTERNAL_0_SELF_ANY_TH',
#     'PRIMER_LEFT_0_SELF_END_TH', 'PRIMER_RIGHT_0_SELF_END_TH', 'PRIMER_INTERNAL_0_SELF_END_TH',
#     'PRIMER_LEFT_0_HAIRPIN_TH', 'PRIMER_RIGHT_0_HAIRPIN_TH', 'PRIMER_INTERNAL_0_HAIRPIN_TH',
#     'PRIMER_LEFT_0_END_STABILITY', 'PRIMER_RIGHT_0_END_STABILITY',
#     'PRIMER_PAIR_0_COMPL_ANY_TH', 'PRIMER_PAIR_0_COMPL_END_TH',
# ]
primer3_col_names = [
    'PRIMER_PAIR_0_PRODUCT_SIZE',
    'PRIMER_PAIR_0_PENALTY',
    'PRIMER_LEFT_0_SEQUENCE', 'PRIMER_RIGHT_0_SEQUENCE',
    'PRIMER_LEFT_0_PENALTY', 'PRIMER_RIGHT_0_PENALTY',
    'PRIMER_LEFT_0_TM', 'PRIMER_RIGHT_0_TM',
    'PRIMER_LEFT_0_GC_PERCENT', 'PRIMER_RIGHT_0_GC_PERCENT',
    'PRIMER_LEFT_0_SELF_ANY_TH', 'PRIMER_RIGHT_0_SELF_ANY_TH',
    'PRIMER_LEFT_0_SELF_END_TH', 'PRIMER_RIGHT_0_SELF_END_TH',
    'PRIMER_LEFT_0_HAIRPIN_TH', 'PRIMER_RIGHT_0_HAIRPIN_TH',
    'PRIMER_LEFT_0_END_STABILITY', 'PRIMER_RIGHT_0_END_STABILITY',
    'PRIMER_PAIR_0_COMPL_ANY_TH', 'PRIMER_PAIR_0_COMPL_END_TH',
]
primer3_col_key = {n: n.replace("PRIMER_", "").replace("_0", "").lower() for n in primer3_col_names}

# Globals
logger = logging.getLogger(__name__)
failure_event = None

def configure_global_logger(args=None, mode="w"):
    global logger
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    stderr_handler = logging.StreamHandler()
    stderr_handler.setLevel(logging.WARNING)
    log_formatter = logging.Formatter('%(levelname)s: %(name)s: %(message)s')
    stderr_handler.setFormatter(log_formatter)
    logger.addHandler(stderr_handler)
    if args is not None:
        if args.log is None:
            if args.log_level is None:
                stderr_handler.setLevel("WARNING")
            else:
                stderr_handler.setLevel(args.log_level)
        else:
            log_file_handler = logging.FileHandler(filename=args.log, mode=mode)
            if args.log_level is None:
                log_file_handler.setLevel("INFO")
            else:
                log_file_handler.setLevel(args.log_level)
            log_file_handler.setFormatter(log_formatter)
            logger.addHandler(log_file_handler)
    return logger


def configure_subprocess_logger(queue):
    global logger
    logger = logging.getLogger(__name__)
    while logger.hasHandlers():
        logger.removeHandler(logger.handlers[0])
    logger.addHandler(QueueHandler(queue))
    logger.setLevel(logging.DEBUG)
    return logger


#class ForkedPdb(pdb.Pdb):
#    """A Pdb subclass that may be used
#    from a forked multiprocessing child
#
#    https://stackoverflow.com/questions/4716533/how-to-attach-debugger-to-a-python-subproccess/23654936#23654936
#    """
#    def interaction(self, *args, **kwargs):
#        _stdin = sys.stdin
#        try:
#            sys.stdin = open('/dev/stdin')
#            pdb.Pdb.interaction(self, *args, **kwargs)
#        finally:
#            sys.stdin = _stdin


def collapse_to_iupac(seqs):
    """Combine sequences into a consensus using IUPAC ambiguity codes
    
    Parameters:
    -----------
    seqs : list of str
        The sequences to combine.
    
    Returns
    -------
    str
        The combined sequence
    """
    seq_lens = [len(x) for x in seqs]
    max_len = max(seq_lens)
    if len(set(seq_lens)) != 1:  # TODO: replace with alignment
        return '-' * max_len
    output = []
    for i in range(max_len):
        column = {s[i] for s in seqs}
        if "*" in column or "N" in column or UNKNOWN_CHAR in column:
            output.append('N')
        else:
            output.append(iupac_key[tuple(sorted(column))])
    return "".join(output)


class GroupedRegion:

    def __init__(self, variants, group, reference, upstream=None, downstream=None):
        """
        Parameters
        ----------
        variants : iterator returning GroupedVariant
            Consecutive variants representing a region.
        group : str
            The group ID for the group to use.
        reference : str
            The reference sequence use to infer the variants
        upstream : deque of GroupedVariant
            The variants upstream (larger reference position) of the region of the interest
        downstream : deque of GroupedVariant
            The variants downstream (smaller reference position) of the region of the interest
       """
        if downstream is None:
            downstream = deque()
        if upstream is None:
            upstream = deque()
        self.variants = deque(variants)
        self.group = group
        self.reference = reference
        self.upstream = upstream
        self.downstream = downstream
        self.type = 'Undetermined'

    @classmethod
    def sliding_window(cls, variants, groups, reference, span, flank=1000):
        """Generate GroupedRegion objects as a sliding window along variants

        Parameters
        ----------
        variants : iterator returning GroupedVariant
            The variants to generate GroupedRegion objects from.
        groups : list of str
            The groups to return sliding windows for
        reference : str
            The file path to the reference sequence use to infer the variants
        span : int
            The number of nucleotides that the variants should span within a group
        flank : int
            The number of variants to keep in the upstream and downstream queues
        """

        def increment(region, flank):
            # Move one variant to spacer and add next to upstream queue
            region.variants.append(region.upstream.popleft())
            # Move from spacer to downstream queue once spacer is too large
            while region.region_length() > span:
                region.downstream.appendleft(region.variants.popleft())
            # Clear downstream queue once it is too large
            while len(region.downstream) > flank:
                region.downstream.pop()

        # Initialize sliding windows
        windows = {}
        for group in groups:
            windows[group] = cls(variants=[], group=group, reference=reference)
        # Read through variants and put into windows
        for index, variant in enumerate(variants):
            for group in groups:
                windows[group].upstream.append(variant)
                if index + 1 >= flank:  # Once the upstream is full
                    increment(windows[group], flank=flank)
                    if len(windows[group].variants) > 0:  # Can be 0 when a single var is larger than the span
                        yield cls(variants=windows[group].variants, group=group, reference=reference,
                                  upstream=windows[group].upstream, downstream=windows[group].downstream)
        # Clear the upstream window
        for index in range(len(next(iter(windows.values())).upstream)):
            for group in groups:
                increment(windows[group], flank=flank)
                if len(windows[group].variants) > 0:  # Can be 0 when a single var is larger than the span
                    yield cls(variants=windows[group].variants, group=group, reference=reference,
                              upstream=windows[group].upstream, downstream=windows[group].downstream)


    @staticmethod
    def _get_reference(ref_path):
        if ref_path is None:
            return None
        else:
            if ref_path.endswith('.gz'):
                handle = gzip.open(ref_path, "rt")
            else:
                handle = open(ref_path)
            reference = list(SeqIO.parse(handle, "fasta"))
            handle.close()
            names = [rec.id for rec in reference]
            reference = dict(zip(names, reference))
            return reference

    def region_length(self):
        """Get the maximum sequence length spanned by variants in a subset
        
        Return
        ------
        int/None
            The number of nucleotides spanned by the variants. If any
            of the samples have variable length alleles, then a single
            length cannot be determined and None is returned.
        """
        # Return 0 if there are no variants
        if len(self.variants) == 0:
            return 0

        # Get length of the reference spanned by these variants
        starts = [x.variant.pos - 1 for x in self.variants]
        ends = [s + x.variant.rlen - 1 for s, x in zip(starts, self.variants)]
        out = max(ends) - min(starts) + 1

        # Apply length changes for each variant to adjust length
        for var in self.variants:
            allele_lens = var.allele_lens(self.group)
            if len(allele_lens) == 0:
                continue
            max_allele_len = max(allele_lens.values())
            ref_len = len(var.variant.ref)
            out += max_allele_len - ref_len

        return out

    @staticmethod
    def _subset_sequence(variants, subset,
                         reference=None,
                         counts=True,
                         hetero=True,
                         unknown=True,
                         min_reads=0,
                         min_geno_qual=0,
                         min_samples=0,
                         consensus=False,
                         conserved=False,
                         wrap='<>',
                         sep=','):
        """For a sample subset, make a sequence representation.
        Parameters
        ----------
        reference : dict of str
            If supplied, non-variant reference positions are also
            returned. made by _get_reference
        counts
            If True, append sample counts to each allele
        hetero
            If False, heterozygous positions are represented by
            IUPAC ambiguity codes instead of being delimited with /
        consensus
            If True, combine all sample alleles into a single allele 
            using IUPAC codes. When a mixture of indel lengths is 
            present, TODO not sure, for now '-' * max indel length
        wrap
            If not None or '', a string composed of the first and last
            character used to border sets of multiple alleles. Has
            no effect if consensus is True.
        sep The character used to separate multiple alleles.
        
        Returns
        -------
        list of str
            A representation of a sequence, with one element per
            variant or single base from the reference sequence.
        """
        # Create rendered text for variant information
        rendered = []
        for variant in variants:
            allele_counts = GroupedVariant._count_genotypes(variant,
                                                            subset=subset,
                                                            hetero=hetero,
                                                            unknown=unknown,
                                                            min_reads=min_reads,
                                                            min_geno_qual=
                                                            min_geno_qual)
            sample_count = GroupedVariant._subset_sample_counts(variant, subset,
                                                                min_reads=min_reads,
                                                                min_geno_qual=
                                                                min_geno_qual)
            if len(allele_counts) == 0:
                allele_counts = {'?': 0}
            if conserved and (len(allele_counts) != 1 or sample_count <
                              min_samples):
                text = None
            else:
                if consensus:
                    text = collapse_to_iupac(allele_counts.keys())
                else:
                    if counts:
                        text_parts = [a + str(c) for a, c in
                                      allele_counts.items()]
                    else:
                        text_parts = allele_counts.keys()
                    text = sep.join(text_parts)
                    if wrap is not None and wrap != '':
                        text = wrap[0] + text + wrap[1]
            rendered.append(text)
        # Insert into reference sequence
        if reference is None:
            return rendered
        else:
            # Get reference sequence spanning variants
            var_pos = [var.pos - 1 for var in variants]
            var_chrom = {var.chrom for var in variants}
            if len(var_chrom) > 1:
                raise ValueError('Variants cannot span multiple chromosomes')
            chrom = list(var_chrom)[0]
            ref_start = min(var_pos)
            ref_end = max(var_pos)
            ref_seq = list(reference[chrom].seq[ref_start:ref_end])
            # Replace reference with each variant
            for var, rend in zip(variants, rendered):
                replace_start = var.pos - 1 - ref_start
                replace_end = replace_start + len(var.ref)
                ref_seq = ref_seq[:replace_start] + [rend] + \
                          ref_seq[replace_end:]
            return ref_seq

    def sequence(self, reference, start, end, group=None, annotate=False):
        """Infer the sequence for the current group based on the reference sequence and variants
        """

        # Find variants within target region
        all_vars = self.downstream + self.variants + self.upstream
        var_starts = [x.variant.pos - 1 for x in all_vars]
        var_ends = [x.variant.pos + x.variant.rlen - 2 for x in all_vars]
        vars_in_range = [v for v, vs, ve in zip(all_vars, var_starts, var_ends) if start <= ve <= end or start <= vs <= end]

        # Check that variants are all on the same chormosome
        if len({x.variant.chrom for x in vars_in_range}) > 1:
            raise ValueError('Variants cannot span multiple chromosomes')
        chrom = self.variants[-1].variant.chrom

        # If there are no variants in the region, return the reference
        if len(vars_in_range) == 0:
            return list(reference[chrom].seq[start:end+1].lower())

        # Insert consensus for each variant in reference sequence
        # for var in vars_in_range:
        #     replace_start = var.variant.pos - 1 - start
        #     replace_end = replace_start + len(var.variant.ref)
        #     if group is None:
        #         consensus = var.variant.ref
        #     else:
        #         alleles = var.allele_counts[group].keys()
        #         if len(alleles) == 0:  # If no data for this position, use N of equal length to reference
        #             consensus = "N" * var.variant.rlen
        #         else:
        #             consensus = collapse_to_iupac(alleles)
        #     if var_upper:
        #         consensus = consensus.upper()
        #     ref_seq = ref_seq[:replace_start] + [consensus] + ref_seq[replace_end:]

        # Sort variants
        var_starts = [v.variant.pos - 1 for v in vars_in_range]
        var_ends = [s + v.variant.rlen - 1 for v, s in zip(vars_in_range, var_starts)]
        var_lens = [e - s + 1 for s, e in zip(var_starts, var_ends)]
        # var_diffs = [l - v.max_allele_len(group) for v, l in zip(vars_in_range, var_lens)]
        vars_in_range = [x for _, x in sorted(zip(var_ends, vars_in_range), key=lambda pair: pair[0])]

        # Adjust start and end positions of sequence to span all variants
        seq_ref_start = min(var_starts + [start])
        seq_ref_end = max(var_ends + [end])

        # Apply variants to reference sequence in reverse order
        out_seq = list(reference[chrom].seq[seq_ref_start:seq_ref_end+1].lower())
        for var in reversed(vars_in_range):
            replace_start = var.variant.pos - 1 - seq_ref_start
            replace_end = replace_start + len(var.variant.ref)
            is_diag_site = any([x is not None for x in var.diagnostic.values()])
            if group is None:
                replacement = var.variant.ref
            else:
                is_diag_for_group = var.diagnostic[group] is not None
                alleles = var.allele_counts[group]
                if len(alleles) == 0:  # If no data for this position, use N of equal length to reference
                    consensus = "N" * var.variant.rlen
                else:
                    consensus = collapse_to_iupac(alleles.keys())
                if annotate:
                    if is_diag_site:
                        replacement = ";".join([k + str(v) for k, v in alleles.items()])
                        if is_diag_for_group:
                            replacement = '<' + replacement + '>'
                    else:
                        replacement = consensus.upper()
                else:
                    if is_diag_for_group:
                        replacement = consensus.upper()
                    else:
                        replacement = consensus.lower()
            if annotate:
                out_seq = out_seq[:replace_start] + [replacement] + out_seq[replace_end:]
            else:
                out_seq = out_seq[:replace_start] + list(replacement) + out_seq[replace_end:]

        # Trim sequence to original start and end points
        if seq_ref_end > end:
            out_seq = out_seq[:len(out_seq) - (seq_ref_end - end)]
        if seq_ref_start < start:
            out_seq = out_seq[start - seq_ref_start:]

        return out_seq

    def conserved(self):
        """For each variant return alleles conserved in a given group

        Returns
        -------
        list of str/None
            Alleles for each variant when there is a conserved allele,
            otherwise None.
        """
        return [x.conserved[self.group] for x in self.variants]

    def diagnostic(self):
        """For each variant return alleles diagnostic for the group

        Returns
        -------
        list of str/None
            Alleles for each variant when there is a conserved allele,
            otherwise None.
        """
        return [x.diagnostic[self.group] for x in self.variants]

    def ref_pos_from_group_offset(self, ref_pos, offset):
        """Get ref index offset by an amount of group-specific sequence"""
        ref_diff_offset = 0
        for v in itertools.chain(reversed(self.downstream), self.variants, self.upstream):
            var_pos_diff = v.variant.pos - 1 - ref_pos
            var_group_offset = var_pos_diff + ref_diff_offset
            if var_group_offset >= offset:
                break
            if var_pos_diff >= 0:
                group_allele_len = v.max_allele_len(self.group)
                ref_allele_len = len(v.variant.ref)
                ref_diff_offset += group_allele_len - ref_allele_len
        return ref_pos + offset - ref_diff_offset


def _parse_reference(ref_path):
    if ref_path is None:
        return None
    else:
        if ref_path.endswith('.gz'):
            handle = gzip.open(ref_path, "rt")
        else:
            handle = open(ref_path)
        reference = list(SeqIO.parse(handle, "fasta"))
        handle.close()
        names = [rec.id for rec in reference]
        reference = dict(zip(names, reference))
        return reference


def parse_primer3_settings(file_path):
    """
    Reads primer3 BoulderIO format, assuming only global settings (starts with PRIMER_),
    and returns a dict in the format used by primer3-py.
    """
    def to_number_if_can(x):
        try:
            if int(float(x)) == float(x) and '.' not in x:
                return int(x)
            else:
                return float(x)
        except ValueError:
            return x

    with open(file_path) as handle:
        options = dict([tuple(l.strip().split('=')) for l in handle.readlines()])
        for opt, val in options.items():
            if ' ' in val or ';' in val:
                val = re.split('[ ;]+', val)
                val = [to_number_if_can(v) for v in val]
                if ',' in val or '-' in val[0]:
                    val = [[to_number_if_can(x) for x in re.split('[,-]+', v)] for v in val]
            elif ',' in val or '-' in val:
                val = re.split('[,\-]+', val)
                val = [to_number_if_can(v) for v in val]
            else:
                val = to_number_if_can(val)
            options[opt] = val
    return options


def run_primer3(template, target_start, target_len,
                options=None,
                tm=(53, 68),
                gc=(40, 70),
                amp_size=(80, 300),
                primer_size=(25, 35),
                max_sec_tm=40,
                gc_clamp=1,
                max_end_gc=4):
    if options is None:
        global_options = {
            'PRIMER_TASK': 'generic',
            'PRIMER_PICK_LEFT_PRIMER': 1,  # 1 == True
            # 'PRIMER_PICK_INTERNAL_OLIGO': 1,  # 1 == True
            'PRIMER_PICK_RIGHT_PRIMER': 1,  # 1 == True
            'PRIMER_LIBERAL_BASE': 1,  # 1 == True
            'PRIMER_OPT_SIZE': mean(primer_size),
            'PRIMER_MIN_SIZE': primer_size[0],
            'PRIMER_MAX_SIZE': primer_size[1],
            # 'PRIMER_INTERNAL_MAX_SIZE': len(crrna_seq),
            'PRIMER_OPT_TM': mean(tm),
            'PRIMER_MIN_TM': tm[0],
            'PRIMER_MAX_TM': tm[1],
            'PRIMER_MIN_GC': gc[0],
            'PRIMER_MAX_GC': gc[1],
            'PRIMER_MAX_POLY_X': 4,  # The maximum allowable length of a mononucleotide repeat
            'PRIMER_MAX_NS_ACCEPTED': 0,  # The maximum number of Ns
            'PRIMER_THERMODYNAMIC_OLIGO_ALIGNMENT': 1,  # 1 == True
            'PRIMER_MAX_SELF_ANY_TH': max_sec_tm,  # describes the tendency of a primer to bind to itself
            'PRIMER_MAX_SELF_END_TH': max_sec_tm,  # 3' binding to itself (primer dimers)
            'PRIMER_PAIR_MAX_COMPL_ANY_TH': max_sec_tm,  # the tendency of the left primer to bind to the right primer
            'PRIMER_PAIR_MAX_COMPL_END_TH': max_sec_tm,  # primer heterodimers
            'PRIMER_MAX_HAIRPIN_TH': max_sec_tm,  # hairpins
            'PRIMER_PRODUCT_SIZE_RANGE': [amp_size],
            'PRIMER_GC_CLAMP': gc_clamp,
            'PRIMER_MAX_END_GC': max_end_gc,
        }
    else:
        global_options = parse_primer3_settings(options)

    p3_output = primer3.bindings.design_primers(
        {
            'SEQUENCE_TEMPLATE': "".join(template),
            # 'SEQUENCE_INTERNAL_OLIGO': "".join(crrna_seq),
            'SEQUENCE_TARGET': [target_start, target_len]
        },
        global_options
    )
    return p3_output


def consv_border_n(group, border_var, nearby_vars, max_offset):
    """Find the maximum size of adjacent conserved reference sequence in terms of reference and group sequence"""
    # Initialize cumulative length differences between reference and variant alleles
    if len(nearby_vars) > 0 and border_var.variant.pos < nearby_vars[0].variant.pos:
        ref_diff_offset = border_var.max_allele_len(group) - len(border_var.variant.ref)
    else:
        ref_diff_offset = 0
    # Initialize distance between reference border var and the start/end of a variant
    ref_diff = 0
    # Search though variants until a non-conserved variant is found
    for nearby_var in nearby_vars:
        # Get info about reference and group
        group_len = nearby_var.max_allele_len(group)
        ref_len = len(nearby_var.variant.ref)
        ref_start = nearby_var.variant.pos
        ref_end = ref_start + ref_len - 1
        # Get distance between reference border var and the start/end of a variant
        if border_var.variant.pos <= ref_start:
            ref_diff = ref_start - border_var.variant.pos
        else:
            ref_diff = border_var.variant.pos - ref_end
        # Check if variant is further away than the maximum offset
        if ref_diff + ref_diff_offset >= max_offset:
            return {"ref": max_offset - ref_diff_offset, "group": max_offset}
        # If variant is not conserved, return pos right before variant
        if nearby_var.conserved[group] is None:
            return {"ref": ref_diff - 1,
                    "group": ref_diff + ref_diff_offset - 1}
        # Keep track of differences in allele length between the group and the reference
        ref_diff_offset += group_len - ref_len
    return {"ref": ref_diff - ref_diff_offset, "group": ref_diff}


def is_nearby_conserved(group, border_var, nearby_vars, max_offset):
    """Check if some number of bases near a variant are conserved """
    ref_diff_offset = 0  # cumulative differences between reference and variant alleles
    for nearby_var in nearby_vars:
        max_allele_len = nearby_var.max_allele_len(group)
        ref_diff_offset += max_allele_len - len(nearby_var.variant.ref)
        var_pos_diff = abs(border_var.variant.pos - nearby_var.variant.pos)
        if var_pos_diff + ref_diff_offset > max_offset:
            return True
        if nearby_var.conserved[group] is None:
            return False


class DiagosticRegion(GroupedRegion):
    """The class for reporting diagnostic regions."""

    def __init__(self,  variants, group, reference, upstream, downstream, p3, crrna_seq, downstream_seq, upstream_seq, temp_range, crrna_range, min_bases):
        super().__init__(variants, group, reference, upstream, downstream)
        self.p3 = p3
        self.downstream_seq = downstream_seq
        self.crrna_seq = crrna_seq
        self.upstream_seq = upstream_seq
        self.temp_range = temp_range
        self.crrna_range = crrna_range
        self.type = "Diagnostic"
        self.min_bases = min_bases

    @staticmethod
    def from_grouped_region(region, p3, crrna_seq, downstream_seq, upstream_seq, temp_range, crrna_range, min_bases):
        return DiagosticRegion(variants=region.variants,
                               group=region.group,
                               reference=region.reference,
                               upstream=region.upstream,
                               downstream=region.downstream,
                               p3=p3,
                               crrna_seq=crrna_seq,
                               downstream_seq=downstream_seq,
                               upstream_seq=upstream_seq,
                               temp_range=temp_range,
                               crrna_range=crrna_range,
                               min_bases=min_bases)

    def left_range(self):
        """Start/stop reference sequence indexes of the left primer."""
        start = self.ref_pos_from_group_offset(ref_pos=self.temp_range[0], offset=self.p3['PRIMER_LEFT_0'][0])
        end = self.ref_pos_from_group_offset(ref_pos=self.temp_range[0], offset=sum(self.p3['PRIMER_LEFT_0']) - 1)
        return [start, end]

    def right_range(self):
        """Start/stop reference sequence indexes of the right primer."""
        start = self.ref_pos_from_group_offset(ref_pos=self.temp_range[0],
                                               offset=self.p3['PRIMER_RIGHT_0'][0] - self.p3['PRIMER_RIGHT_0'][1] + 1)
        end = self.ref_pos_from_group_offset(ref_pos=self.temp_range[0],
                                             offset=self.p3['PRIMER_RIGHT_0'][0])
        return [start, end]

    def missing_samples(self):
        diag_vars = [var for var in self.variants if var.diagnostic[self.group] is not None]
        return {id for var in diag_vars for ids in var.missing_samp_ids.values() for id in ids}

    # def crrna_range(self):
    #     """Start/stop reference sequence indexes of the crRNA."""
    #     # start = self.ref_pos_from_group_offset(ref_pos=self.temp_range[0], offset=self.p3['PRIMER_INTERNAL_0'][0])
    #     # end = self.ref_pos_from_group_offset(ref_pos=self.temp_range[0], offset=sum(self.p3['PRIMER_INTERNAL_0']))
    #     # return [start, end]
    #     return self.crrna_range


def find_diag_region(variants,
                     groups,
                     reference=None,
                     nontarget=None,
                     primer3=False,
                     min_vars=1,
                     min_bases=1,
                     min_samp_prop=0.9,
                     min_samples=5,
                     min_reads=5,
                     min_geno_qual=30,
                     min_map_qual=40,
                     min_var_qual=10,
                     min_freq=0.1,
                     crrna_len=28,
                     tm = (53, 68),
                     gc = (40, 70),
                     amp_size=(80, 300),
                     primer_size=(25,35),
                     max_sec_tm = 40,
                     gc_clamp=1,
                     max_end_gc=4,
                     var_location=(4,16),
                     force=False):
    """Find regions with diagnostic variants
    
    Return information on regions that contain variants diagnostic for a subset
    of samples, optionally surrounded by conserved regions where primers can be
    designed.
    
    Parameters
    ----------
    variants : an iterable returning variants or str
        A series of consecutive variants in which to identify diagnostic
        clusters of variants.
    groups : dict of list/tuple of str
        The sample IDs for each group to find diagnostic clusters for,
        named by group.
    nontarget : list of str, optional
        The sample IDs that should be distinct from those in `groups`, but
        are otherwise not of interest. `min_samples` does not apply to these.
    reference :
        The path to a FASTA file. Required if Primer3 is to be used.
    primer3 : bool, optional
        If `True`, run Primer3 on each potential cluster of diagnostic
        variants and only return results for which primers can be found
        in adjacent conserved sequence.
    min_vars : int, optional
        The minimum number of diagnostic variants. Note that a single
        indel variant can represent multiple diagnostic nucleotides.
        See `min_bases` to filter based on number of nucleotides.
    min_bases : int, optional
        The minimum ngroupsumber of diagnostic nucleotides. Note that a single
        indel variant can represent multiple diagnostic nucleotides.
        See `min_vars` to filter based on number of variants.
    min_samp_prop : float, optional
        The minimum proportion of samples that must contain a diagnositc variant in each group
    min_samples : int, optional
        The minimum number of samples that must represent each group in
        `groups` for a given variants. Samples must pass the `min_reads` 
        and `min_geno_qual` filters to count towards this minimum.
    min_reads : int, optional
        The minimum number of reads a sample have at the location of a
        given variant.
    min_geno_qual : int, optional
        The minimum genotype quality score (phred scale). This corresponds
        to the per-sample output in the VCF encoded as "GQ".
    min_map_qual : int, optional
        The minimum root square mean mapping quality of reads supporting a
        variant (phred scale). This corresponds to the per-variant
        output in the VCF encoded as "MQ".
    min_var_qual : int, optional
        Phred-scaled quality score for the assertion that a variant exists.
        This corresponds to the per-variant output in the VCF encoded as the "QUAL" column.
    min_freq : float, optional
        The minimum proportion of reads an allele must have to be considered real.
        This is meant to counter sequencing errors.
        If not set, the genotypes called in the VCF are used without considering read depth.
    crrna_len: int, optional
        The length of the spacer sequence that the crRNA will bind to and
        the primers will amplify.
    offset_left : int, optional
        The minimum number of bases on the 3' end of the spacer before
        the first diagnostic SNP.
    offset_right : int, optional
        The number of bases on the 5' end after the last diagnostic SNP.
        
    Yields
    ------
    object?
        information on regions that contain variants diagnostic for a
        subset of samples. TBD.
    """
    offset_left = var_location[0] - 1
    offset_right = crrna_len - var_location[1]
    window_width = crrna_len - offset_right - offset_left

    vcf_reader = GroupedVariant.from_vcf(variants, groups,
                                         min_samp_prop=min_samp_prop,
                                         min_samples=min_samples,
                                         min_reads=min_reads,
                                         min_geno_qual=min_geno_qual,
                                         min_freq=min_freq,
                                         min_map_qual=min_map_qual,
                                         min_var_qual=min_var_qual,
                                         force=force)
    windower = GroupedRegion.sliding_window(vcf_reader,
                                            groups=groups.keys(),
                                            reference=reference,
                                            span=window_width,
                                            flank=amp_size[1])
    for region in windower:

        # Are there enough diagnostic variants?
        is_diag = [x is not None for x in region.diagnostic()]
        n_diag_var = sum(is_diag)
        if n_diag_var < min_vars:
            region.type = 'Undiagnostic'
            yield region
            continue

        # Are all the variants in the crRNA conserved?
        if any([x is None for x in region.conserved()]):
            region.type = 'Unconserved'
            yield region
            continue

        # If there is only one diagnostic variants, is it on the right?
        # NOTE: this might miss some valid regions when a conservered variant is a few bp more to the right
        if n_diag_var == 1 and is_diag[-1] is False:
            region.type = 'Misplaced'
            yield region
            continue

        # # Is there conserved sequence adjacent to the variants to fit the crRNA?
        # region_len = region.region_length()
        # overhang_left = spacer_len - offset_right - region_len
        # if not is_nearby_conserved(region.group, region.variants[-1], region.upstream, offset_right + 1):
        #      region.type = 'Unconserved'
        #      yield region
        # if not is_nearby_conserved(region.group, region.variants[0], region.downstream, overhang_left + 1):
        #      region.type = 'Unconserved'
        #      yield region

        # Is there conserved adjacent sequence for the crRNA?
        overhang_left = crrna_len - region.region_length() - offset_right
        overhang_right = offset_right
        overhang_len_up = consv_border_n(group=region.group,
                                         border_var=region.variants[-1],
                                         nearby_vars=region.upstream,
                                         max_offset=overhang_right)
        overhang_len_dn = consv_border_n(group=region.group,
                                         border_var=region.variants[0],
                                         nearby_vars=region.downstream,
                                         max_offset=overhang_left)
        if overhang_len_up['group'] < offset_right or overhang_len_dn['group'] < overhang_left:
            region.type = 'Unconserved'
            yield region
            continue

        # Is there enough adjacent conserved regions to design primers?
        consv_len_up = consv_border_n(group=region.group,
                                      border_var=region.variants[-1],
                                      nearby_vars=region.upstream,
                                      max_offset=amp_size[1])
        consv_len_dn = consv_border_n(group=region.group,
                                      border_var=region.variants[0],
                                      nearby_vars=region.downstream,
                                      max_offset=amp_size[1])
        if consv_len_up["group"] - overhang_len_up['group'] < 30:  #TODO base on primer3 primer size
            region.type = 'Unconserved'
            yield region
            continue
        if consv_len_dn["group"] - overhang_len_dn['group'] < 30:  #TODO base on primer3 primer size parameters
            region.type = 'Unconserved'
            yield region
            continue


        # Are there enough diagnostic bases?
        start_crrna_ref = region.variants[0].variant.pos - 1 - overhang_len_dn['ref']
        end_crrna_ref = region.variants[-1].variant.pos - 1 + overhang_len_up['ref']
        crrna_seq = region.sequence(reference=reference, start=start_crrna_ref, end=end_crrna_ref, group=region.group)
        crrna_nontaget_seqs = [region.sequence(reference=reference, start=start_crrna_ref, end=end_crrna_ref, group=g) for g in groups if g is not region.group]
        edit_dists = [distance.edit_distance("".join(crrna_seq), "".join(seq)) for seq in crrna_nontaget_seqs]
        min_dist = min(edit_dists)
        if min_dist < min_bases:
            region.type = 'Undiagnostic'
            yield region
            continue

        start_tmp_ref = region.variants[0].variant.pos - 1 - consv_len_dn['ref']
        end_tmp_ref = region.variants[-1].variant.pos - 1 + consv_len_up['ref']

        # Are there no overlapping variants TODO: fix so this is not needed
        #all_vars = region.downstream + region.variants + region.upstream
        #var_starts = [x.variant.pos - 1 for x in all_vars]
        #var_ends = [x.variant.pos + x.variant.rlen - 2 for x in all_vars]
        #vars_in_range = [v for v, vs, ve in zip(all_vars, var_starts, var_ends) if start_tmp_ref <= ve <= end_tmp_ref or start_tmp_ref <= vs <= end_tmp_ref]
        #overlapping = False
        #for v1, v2 in zip(list(vars_in_range)[:-1], list(vars_in_range)[1:]):
        #    if v1.variant.stop > v2.variant.start:
        #        region.type = 'Overlapping'
        #        overlapping = True
        #        break
        #if overlapping:
        #    yield region
        #    continue

        # Run primer3 on group-specific template
        downstream_seq = region.sequence(reference=reference, start=start_tmp_ref, end=start_crrna_ref-1, group=region.group)
        upstream_seq = region.sequence(reference=reference, start=end_crrna_ref+1, end=end_tmp_ref, group=region.group)

        template_seq = downstream_seq + crrna_seq + upstream_seq
        start_crrna_tmp = len(downstream_seq)
        end_crrna_tmp = start_crrna_tmp + len(crrna_seq) - 1

        p3_out = run_primer3(template_seq, target_start=start_crrna_tmp, target_len=len(crrna_seq),
                             tm=tm,
                             gc=gc,
                             amp_size=amp_size,
                             primer_size=primer_size,
                             max_sec_tm=max_sec_tm,
                             gc_clamp=gc_clamp,
                             max_end_gc=max_end_gc)

        if p3_out['PRIMER_PAIR_NUM_RETURNED'] == 0:
            region.type = 'No primers'
            yield region
            continue

        # Return data for the diagnostic region
        region.type = 'Diagnostic'
        output = DiagosticRegion.from_grouped_region(region, p3=p3_out, crrna_seq=crrna_seq, downstream_seq=downstream_seq, upstream_seq=upstream_seq,
                                                     temp_range=[start_tmp_ref, end_tmp_ref], crrna_range=[start_crrna_ref, end_crrna_ref], min_bases=min_dist)

        yield output


def parse_command_line_args():
    # Parse command line arguments
    logger.debug('Parsing command line arguments')
    parser = argparse.ArgumentParser(
        description='Find regions where there are conserved variants for each group that are not found in other groups.')
    parser.add_argument('metadata', type=str, metavar='METADATA',
                        help='A CSV file containing data with one row per sample. Two columns are required: one which contains the same sample IDs used in the VCF file, and one which identifies which group each sample belongs to. See --sample_col and --group_col for default values and how to specify custom column names.')
    parser.add_argument('reference', type=str, metavar='REFERENCE',
                        help='The reference file used to make the VCF input. If supplied, the sequence of the region containing each conserved variant is returned with the output.')
    parser.add_argument('--vcf', type=str, default="-", metavar='PATH',
                        help='A VCF file containing variant data for the samples grouped by the `metadata` file. If not supplied, VCF data will be read from stanard input (stdin) using a single core. (default: read from stdin)')
    parser.add_argument('--sample_col', type=str, default="sample_id", metavar='TEXT',
                        help='The names of column in the metadata containing sample IDs. (default: %(default)s)')
    parser.add_argument('--group_col', type=str, default="group", metavar='TEXT',
                        help='The names of column in the metadata containing group IDs. Samples with the same groups ID will represent that group. (default: %(default)s)')
    parser.add_argument('--index', type=str, metavar='PATH',
                        help='The path to an tabix index file for the VCF file. If not supplied, a file with the same name as the VCF file with .tbi/.csi appended will be searched for. If that is not found, an index file will be created in the same directory as the VCF file.')
    parser.add_argument('--groups', type=str, nargs="+", metavar='TEXT',
                        help='One or more groups that are to be distinguished by variants. These should match the values of the column specified by --group_col in the metadata file. (default: use all groups)')
    parser.add_argument('--out_csv', type=str, metavar='PATH',
                        help='The output file to create. If not supplied, results will be printed to the screen (standard out). (default: print to stdout)')
    parser.add_argument('--out_align', type=str, metavar='PATH',
                        help='A file path to print human-readable alignments of diagnostic regions. (default: do not output)')
    parser.add_argument('--chroms', type=str, nargs="+", metavar='TEXT',
                        help='One or more chromosomes (reference fasta headers) to restrict the search to. (default: use all chromosomes)')
    parser.add_argument('--pos', type=int, nargs=2, metavar='INT', default=None,
                        help='The range of indexes (1-based) of positions in the reference sequences to search. (default: search whole chromosome)')
    parser.add_argument('--min_samples', type=int, default=3, metavar='INT',
                        help='The number of samples with data passing quality filters each group must have for a given variant. (default: %(default)s)')
    parser.add_argument('--min_samp_prop', type=float, default=0.9, metavar='PROP',
                        help='The minimum proportion of samples that must contain a diagnositc variant in each group. (default: %(default)s)')
    parser.add_argument('--min_reads', type=int, default=10, metavar='INT',
                        help='The number of reads a variant must be represented by in a given sample for the data of that sample to be considered. This corresponds to the per-sample GATK output in the VCF encoded as "DP". (default: %(default)s)')
    parser.add_argument('--min_geno_qual', type=int, default=40, metavar='INT',
                        help='The minimum genotype quality score (phred scale). This corresponds to the per-sample GATK output in the VCF encoded as "GQ". (default: %(default)s)')
    parser.add_argument('--min_var_qual', type=int, default=10, metavar='INT',
                        help='Phred-scaled quality score for the assertion that a variant exists. This corresponds to the per-variant output in the VCF encoded as the "QUAL" column. (default: %(default)s)')
    parser.add_argument('--min_freq', type=float, default=0.1, metavar='PROP',
                        help='The minimum proportion of reads an allele must have to be considered real. This is meant to counter sequencing errors. (default: %(default)s)')
    parser.add_argument('--min_map_qual', type=int, default=40, metavar='INT',
                        help='The minimum root square mean mapping quality of reads supporting a variant (phred scale). This corresponds to the per-variant output in the VCF encoded as "MQ" in the "INFO" column. (default: %(default)s)')
    parser.add_argument('--min_bases', type=int, default=1, metavar='INT',
                        help='The minimum number of bases distinguishing the target group from other groups. (default: %(default)s)')
    parser.add_argument('--cores', type=int, default=1, metavar='INT',
                        help='The number of processors to use for parallel processing. (default: %(default)s)')
    parser.add_argument('--log', type=str, metavar='PATH',
                        help='The location to save a log file containing information, warnings, and errors. (default: print to screen via stderr)')
    parser.add_argument('--log_level', type=str, choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
                        help='The minimum importance level of messages to print to the log. (default: INFO if the log is being saved to a file with --log, otherwise WARNING)')
    parser.add_argument('--var_location', type=int, nargs=2, metavar='INT', default=[6, 14],
                        help="The range of acceptable locations for diagnostic variants in the crRNA, measured from the 5' end. (default: %(default)s)")
    parser.add_argument('--crrna_len', type=int, default=28, metavar='INT',
                        help='Then length of the crRNA. (default: %(default)s)')
    parser.add_argument('--tm', type=int, nargs=2, metavar='INT', default=[53, 68],
                        help='The minimum and maximum melting temperature when searching for primers. (default: %(default)s)')
    parser.add_argument('--gc', type=int, nargs=2, metavar='INT', default=[40, 70],
                        help='The minimum and maximum GC percentage when searching for primers. (default: %(default)s)')
    parser.add_argument('--amp_size', type=int, nargs=2, metavar='INT', default=[70, 150],
                        help='The minimum and maximum size of the amplicon when searching for primers. (default: %(default)s)')
    parser.add_argument('--primer_size', type=int, nargs=2, metavar='INT', default=[25, 35],
                        help='The minimum and maximum size of the primers. (default: %(default)s)')
    parser.add_argument('--max_sec_tm', type=int, default=40, metavar='INT',
                        help='The maximum melting temperature of any secondary structures when searching for primers, including hetero/homo dimers and hairpins. (default: %(default)s)')
    parser.add_argument('--gc_clamp', type=int, default=1, metavar='INT',
                        help="Require the specified number of consecutive Gs and Cs at the 3' end of both the left and right primer. (default: %(default)s)")
    parser.add_argument('--max_end_gc', type=int, default=4, metavar='INT',
                        help="The maximum number of Gs or Cs allowed in the last five 3' bases of a left or right primer. (default: %(default)s)")
    parser.add_argument('--force', action='store_true', default=False,
                        help="Try to continue executing when the program would typically quit.")
    args = parser.parse_args()

    return args


def read_vcf_contigs(path, reference, index=None, chunk_size=100000, flank_size=1000, contig_subset=None, pos_subset=None):
    """Supply iterators for contigs/scaffolds/chormosomes in a VCF file.

    Parameters
    ----------
    path : str
        The path to the VCF file
    index : str, optional
        The path to the index file for the VCF
    """
    # If reading from stdin, return None
    if path == "-":
        return [None]
    # Get path to index file or create it
    if index is None:
        tbi_path = path + '.tbi'
        csi_path = path + '.csi'
        if os.path.isfile(tbi_path):
            index = tbi_path
        elif os.path.isfile(csi_path):
            index = csi_path
        else:
            logger.info(f'Creating index file  "{tbi_path}"')
            pysam.tabix_index(path, preset='vcf', keep_original=True, force=True)
            index = path + '.gz.tbi'
            path = path + '.gz'
    # Reduce chunk size if too big for custom position range
    if pos_subset is not None:
        pos_length = max(pos_subset) - min(pos_subset) + 1
        if pos_length < chunk_size:
            chunk_size = pos_length
    # Split contigs into chunks for processing
    index_handle = pysam.TabixFile(filename=path, index=index)
    output = []
    for contig in index_handle.contigs:
        if contig_subset is not None and contig not in contig_subset:
            continue
        if pos_subset is None:
            search_start = 0
            search_end = len(reference[contig])
        else:
            search_start = min(pos_subset) - 1
            search_end = max(pos_subset) - 1
        for start in range(search_start, search_end, chunk_size):
            end = start + chunk_size + flank_size
            if start > flank_size:
                start -= flank_size
            output.append({'contig': contig, 'start': start, 'end': end})

    return output


def _format_p3_output(p3_out):
    """Reformat data for best primer pair for CSV output"""
    return {primer3_col_key[n]: p3_out[n] for n in primer3_col_names}


def _render_header(stream, out_sep=','):
    print("group", "chrom", "n_diag", "fwd_start", "fwd_end", "rev_start",
          "rev_end", "crrna_start", "crrna_end", "seq_start", "seq_end",
          *[primer3_col_key[n] for n in primer3_col_names],
          sep=out_sep, file=stream)


@contextmanager
def stream_writer(file_path=None, default_stream=sys.stdout):
    file_handle = default_stream if file_path is None else open(file_path, "w")
    yield file_handle
    if file_path is not None:
        file_handle.close()


def _format_for_csv(region, reference, groups):
    fwd_range = region.left_range()
    rev_range = region.right_range()
    crrna_range = region.crrna_range
    temp_range = region.temp_range

    group = region.group
    chrom = region.variants[0].variant.chrom
    # n_diag = sum([x is not None for x in region.diagnostic()])
    n_diag = region.min_bases

    fwd_start = fwd_range[0] + 1  # The +1 goes from 0-based in 1-based indexing
    fwd_end = fwd_range[1] + 1
    rev_start = rev_range[0] + 1
    rev_end = rev_range[1] + 1
    crrna_start = crrna_range[0] + 1
    crrna_end = crrna_range[1] + 1
    seq_start = region.temp_range[0] + 1
    seq_end = region.temp_range[1] + 1

    # Make unannotated sequences of target group
    def format_seq(start, end):
        out = region.sequence(start=start, end=end,
                              reference=reference, group=region.group, annotate=False)
        return "".join(out)

    seq_adj_left = format_seq(start=temp_range[0], end=fwd_range[0] - 1)
    seq_primer_left = format_seq(start=fwd_range[0], end=fwd_range[1])
    seq_amp_left = format_seq(start=fwd_range[1] + 1, end=crrna_range[0] - 1)
    seq_crrna = format_seq(start=crrna_range[0], end=crrna_range[1])
    seq_amp_right = format_seq(start=crrna_range[1] + 1, end=rev_range[0] - 1)
    seq_primer_right = format_seq(start=rev_range[0], end=rev_range[1])
    seq_adj_right = format_seq(start=rev_range[1] + 1, end=temp_range[1])

    missing = region.missing_samples()
    missing_samp_ids = ';'.join(missing)
    missing_count = len(missing)


    # Modify these values to change columns and their names:
    output = {
        "region_id": f'{chrom}:{fwd_start}-{rev_end}',
        "group": group,
        "chrom": chrom,
        "n_diag": n_diag,
        "n_missing": missing_count,
        "reg_from": seq_start,
        "reg_to": seq_end,
        "diag_from": crrna_start,
        "diag_to": crrna_end,
        "fwd_from": fwd_start,
        "fwd_to": fwd_end,
        "rev_from": rev_start,
        "rev_to": rev_end,
        "missing_samp_ids": missing_samp_ids,
        "seq_adj_left": seq_adj_left,
        "seq_primer_fwd": seq_primer_left,
        "seq_inter_left": seq_amp_left,
        "seq_diag": seq_crrna,
        "seq_inter_right": seq_amp_right,
        "seq_primer_rev": seq_primer_right,
        "seq_adj_right": seq_adj_right
    }
    output.update(_format_p3_output(region.p3))
    # output.update(group_seqs)

    return output


def _print_alignment(region, reference, groups):

    fwd_range = region.left_range()
    rev_range = region.right_range()
    crrna_range = region.crrna_range

    # Format sequences for each group
    def format_seq(group, start, end):
        out = region.sequence(start=start, end=end,
                              reference=reference, group=group, annotate=True)
        return out
    group_seqs = {g: format_seq(g, start=fwd_range[0], end=rev_range[1]) for g in groups}
    ref_seq = format_seq(None, start=fwd_range[0], end=rev_range[1])

    # Format primer and crrna sequences
    def format_oligo(start, end):
        out = region.sequence(start=start, end=end,
                              reference=reference, group=region.group, annotate=False)
        return "".join(out)
    seq_primer_left = format_oligo(start=fwd_range[0], end=fwd_range[1])
    seq_primer_right = format_oligo(start=rev_range[0], end=rev_range[1])
    seq_crrna = format_oligo(start=crrna_range[0], end=crrna_range[1])
    # oligos = {
    #     seq_primer_left: 0,
    #     seq_crrna: crrna_range[0] - fwd_range[0],
    #     seq_primer_right: rev_range[0] - fwd_range[0]
    # }
    oligos = [
        Annotation(name="Left primer", seq=seq_primer_left, start=0),
        Annotation(name="crRNA", seq=seq_crrna, start=crrna_range[0] - fwd_range[0]),
        Annotation(name="Right primer", seq=seq_primer_right, start=rev_range[0] - fwd_range[0])
    ]

    chrom = list(region.reference.keys())[0]
    start = fwd_range[0] + 1
    end = rev_range[1] + 1
    group = region.group
    output = [f"## {chrom}:{start}-{end} is diagnostic for {region.group}\n"]
    try:
        output += render_variant(seqs=group_seqs, ref=ref_seq, p3=region.p3, groups=groups, annots=oligos)
    except (IndexError, TypeError) as error:
        output += ["CANNOT PRINT ALIGNMENT WITH OVERLAPPING INDELS"]
        logger.info(f"Failed to print alignment of {chrom}:{start}-{end} diagnostic for {region.group} due to overlapping indels.")
    output += ['\n']

    return output


def report_diag_region(vcf_path, contig, groups, reference, args, **kwargs):
    if contig is None:  # If reading from stdin
        logger.info(f"Reading VCF data from standard input (stdin).")
        variants = pysam.VariantFile("-")
    else:
        logger.info(f"Starting scan of contig {contig['contig']}:{contig['start']}-{contig['end']}")
        vcf_handle = pysam.VariantFile(vcf_path)
        variants = vcf_handle.fetch(contig['contig'], start=contig['start'], end=contig['end'])
    stats = defaultdict(int)
    undiag_count = 0 # How many variants are not diagnositic and have therefore not been reported yet
    update_interval = 1000 # How often stats on non-diagnostic variants are reported
    for region in find_diag_region(variants, groups, reference, **kwargs):
        if failure_event is not None and failure_event.is_set():
            logger.critical("Error detected in other worker process. Ending this process too.")
            return None
        stats[region.type] += 1
        if region.type == "Diagnostic":
            output = _format_for_csv(region, reference, groups)
            if args.out_align is None:
                alignment = None
            else:
                alignment = _print_alignment(region, reference, groups)
            yield {'result': output, 'stats': stats, 'alignment': alignment}
            stats = defaultdict(int)
        else:
            undiag_count += 1
        if undiag_count >= update_interval:
            yield {'result': None, 'stats': stats, 'alignment': None}
            undiag_count = 0
            stats = defaultdict(int)
    return None




class ResultWriter:

    def __init__(self, output_stream, groups, align_path=None):
        self.result_header_printed = False
        self.stat_header_printed = False
        self.stats = defaultdict(int)
        self.output_stream = output_stream
        self.stat_names = ['Undiagnostic', 'Unconserved', 'No primers']
        self.variant_counts = {s: 0 for s in self.stat_names}
        self.groups = groups
        self.group_counts = {g: 0 for g in groups}
        self.align_path = align_path
        if align_path is not None:
            self.out_align = open(align_path, "w")

    def print_result(self, result):
        if not self.result_header_printed:
            print(*result.keys(), sep=',', file=self.output_stream, flush=True)
            self.result_header_printed = True
        print(*result.values(), sep=',', file=self.output_stream, flush=True)

    def print_stats_header(self):
        max_nchar = max([len(n) for n in self.stat_names + self.groups])
        header_parts = [n.ljust(max_nchar) for n in self.stat_names + self.groups]
        print('| '.join(header_parts), file=sys.stderr)

    def print_status(self, end_line=False):
        if not self.stat_header_printed:
            self.print_stats_header()
            self.stat_header_printed = True
        max_nchar = max([len(n) for n in self.stat_names + self.groups])
        var_info = [str(self.variant_counts[n]).ljust(max_nchar) for n in self.stat_names]
        group_info = [str(self.group_counts[n]).ljust(max_nchar) for n in self.groups]
        print('| '.join(var_info + group_info), file=sys.stderr, end='\n' if end_line else '\r')

    def update_stats(self, output):
        if output['result'] is not None:
            self.group_counts[output['result']['group']] += 1
        for stat, count in output['stats'].items():
            if stat in self.variant_counts:
                self.variant_counts[stat] += count

    def write_alignment(self, output):
        if self.align_path is not None:
            self.out_align.writelines([x + '\n' for x in output] + ['\n'])

    def write(self, output):
        if output['result'] is not None:
            self.print_result(output['result'])
            self.write_alignment(output['alignment'])
        self.update_stats(output)
        self.print_status()

    def finish(self):
        print('', file=sys.stderr) #move line down so stats are not covered


def mp_worker(queue, log_queue, args, contig, groups, reference, **kwargs):
    configure_subprocess_logger(log_queue)
    try:
        for result in report_diag_region(args.vcf, contig, groups, reference, args, **kwargs):
            queue.put(result)
    except BaseException as error:
        logger.exception(f"Error encountered while scanning contig {contig['contig']}:{contig['start']}-{contig['end']}:")
        failure_event.set()
        raise error
    return None


def mp_listener(result_queue, log_queue, args):
    '''listens for output on the queue and writes to a file. '''
    global logger
    logger = configure_global_logger(args, mode="a")
    with stream_writer(args.out_csv, sys.stdout) as output_stream:
        writer = ResultWriter(output_stream, args.groups, align_path=args.out_align)
        while True:
            # Write one result returned by workers
            try:
                output = result_queue.get(block=False, timeout=0.1)
                if output == "kill":
                    break
                writer.write(output)
            except queue.Empty:
                pass

            # Write all logs returned by workers
            while True:
                try:
                    logs = log_queue.get(block=False)
                    # logger = logging.getLogger(logs.name)
                    logger.handle(logs)
                except queue.Empty:
                    break
    writer.finish()
    total_vars = sum(writer.variant_counts.values()) + sum(writer.group_counts.values())
    logger.info("Total variants scanned: " + str(total_vars))


def mp_worker_init(event):
    global failure_event
    failure_event = event


def run_all():

    # Parse command line arguments
    args = parse_command_line_args()

    # Log record of parsed parameters
    global logger
    logger = configure_global_logger(args)
    lines = [f"    {k : <15}: {v}" for k, v in vars(args).items() if v is not None]
    logger.info("\n".join(["Parameters used:"] + lines))

    # Prepare input data
    reference = _parse_reference(args.reference)
    groups = _parse_group_data(args.metadata, groups=args.groups, sample_col=args.sample_col,
                               group_col=args.group_col, min_samples=args.min_samples)
    contigs = read_vcf_contigs(args.vcf, reference=reference, chunk_size=100000, flank_size=1000,
                               contig_subset=args.chroms, pos_subset=args.pos) #TODO base on amplicon size, need to add to args
    gz_path = args.vcf + '.gz'
    if os.path.isfile(gz_path):
        args.vcf = gz_path
    search_arg_names = ('min_samples', 'min_reads', 'min_geno_qual', 'min_map_qual', 'min_var_qual', 'min_freq',
                        'min_samp_prop', 'var_location', 'crrna_len', 'tm', 'gc', 'primer_size', 'amp_size',
                        'max_sec_tm', 'min_bases', 'gc_clamp', 'max_end_gc', 'force')
    search_args = {k: v for k, v in vars(args).items() if k in search_arg_names}

    if args.vcf != "-" and args.cores > 1:
        # Prepare for multiprocessing
        logger.debug('Preparing for multiprocessing')
        mp.set_start_method('spawn')
        manager = mp.Manager()
        failure_event = manager.Event()
        queue = manager.Queue()
        log_queue = manager.Queue()
        pool = mp.Pool(args.cores + 1, initializer=mp_worker_init, initargs=(failure_event,))
        watcher = pool.apply_async(mp_listener, args=(queue, log_queue, args))

        # Set main process logger to submit to the listener process instead
        configure_subprocess_logger(log_queue)

        # Spawn processes
        logger.debug('Spawning multiprocessing workers')
        jobs = []
        for contig in contigs:
            job = pool.apply_async(mp_worker,
                                   args=(queue, log_queue, args, contig, groups, reference),
                                   kwds=search_args)
            jobs.append(job)

        # End multiprocessing
        logger.debug('Waiting for multiprocessing workers to finish')
        for job in jobs:
            job.get()
        queue.put('kill')
        pool.close()
        pool.join()

        # Reset logger to use main process
        logger = configure_global_logger(args, mode='a')
    else:
        logger.debug('Running code on a single core')
        with stream_writer(args.out_csv, sys.stdout) as output_stream:
            writer = ResultWriter(output_stream, args.groups, align_path=args.out_align)
            for contig in contigs:
                for result in report_diag_region(args.vcf, contig, groups, reference, args,
                                                 **search_args):
                    writer.write(result)
            writer.finish()
        total_vars = sum(writer.variant_counts.values()) + sum(writer.group_counts.values())
        logger.info("Total variants scanned: " + str(total_vars))


def main():
    # import cProfile; cProfile.runctx('run_all()', globals=globals(), locals=locals(), sort='cumulative')
    run_all()


if __name__ == "__main__":
    # Test command:
    # python -m diagvar.find_diag_region test_data/test_metadata.csv test_data/unfilt_allscafs_n666.vcf.gz --groups NA1 NA2 EU1 EU2 --reference test_data/PR-102_v3.1_s0001.fasta --out test.csv
    main()

# if __name__ == "__main__":
#     vcf = pysam.VariantFile('test_data/unfilt_allscafs_n666.vcf.gz')
#     groups = _read_group_data('test_data/test_metadata.csv')
#     groups = {g: v for g, v in groups.items() if g in ["NA1", "NA2", "EU1", "EU2"]}
#     ref = 'test_data/PR-102_v3.1.fasta'
#     find_diag_region(vcf,
#                      groups=groups,
#                      reference=ref)




import math
import tempfile
import itertools
from pathlib import Path
from .Amplicon import Amplicon, ConservedEndAmplicons


def prettyTime(time):
    """ Converts seconds into a readable minutes-seconds output format.

    Parameters
    ----------
    time : float
        Input time in seconds to convert

    Returns
    -------
    str
        A pretty version of the time to print
    """
    if time < 60:
        # Print seconds to 2 decimal places
        return f"{time:.2f} seconds"
    else:
        # Determine the number of minutes and remaining seconds
        minutes = int(time / 60)
        seconds = math.ceil(time - 60*minutes)
        # Determine if minutes and seconds are plural
        pl_min = "s" if (minutes > 1) else ""
        pl_sec = "s" if (seconds > 1) else ""
        return f"{minutes} minute{pl_min} and {seconds} second{pl_sec}"


def basename(filename):
    """Return the basename of a compressed fasta file.

    Parameters
    ----------
    filename : str
        The name of a file

    Returns
    -------
    str
        The basename of the file without fasta and compression endings

    """
    # Use Path library to get the filename from a full directory path and split
    filename = Path(filename).name.split('.')
    # Iterate and remove extensions from list
    omit = ["gz", "bz2", "fna", "fasta", "fa", "ffn", "frn"]
    while filename[-1] in omit:
        filename.pop()
    # Return what remains as the result
    return '.'.join(filename)


def simplename(filename):
    """Return the basename of a fasta file with all extensions removed.

    Parameters
    ----------
    filename : str
        The name of a file

    Returns
    -------
    str
        The basename of the file with no extensions

    """
    # Return what remains as the result
    return basename(filename).split('.')[0]


def tmpFile(workdir=None):
    """ Create a temporary file that won't be deleted on close.

    Parameters
    ----------
    workdir : str, optional
        Directory to store the newly created temporary file

    Returns
    -------
    str
        Filename of temporary file created
    """
    # Get name of working directory
    if workdir is None:
        workdir = tempfile.gettempdir()

    # Make a temporary file in a context s.t. it is closed automatically
    filename = ""
    with tempfile.NamedTemporaryFile(dir=workdir, delete=False) as fptr:
        # Get name parameter
        filename = fptr.name
    # Return filename
    return filename


def fileSize(fptr):
    """ Determines the length of an open file in bytes.

    This function determines the length of a file by moving the file pointer
    to the end of the file to get the position of the last byte. The pointer
    is returned to its original location unmodified.

    Parameters
    ----------
    fptr : open file-ptr like object

    Results
    -------
    int
        Size of the open file in bytes

    """
    # Get current pos
    origin = fptr.tell()

    # Move ptr to end of file
    fptr.seek(0, 2)

    # Get byte value as size
    file_size = fptr.tell()

    # Move ptr back to start and return size
    fptr.seek(origin)
    return file_size


def splitFilePtrs(filename, num_sections):
    """ Split a file into roughly equal sections.

    This function attempts to split a file into equally sized sections and
    return the start and end byte values for each section. The sections are
    only roughly equal as line lengths may differ and the file may not be
    equally divisable. The general algorithm is as follows:
        1) Open file for reading
        2) Get file size in bytes from open file pointer
        3) Determine the approximate size of each section in bytes
        4) Iterate to get official start and end bytes under the conditions:
            a) start byte is the beginning of a line


    Parameters
    ----------
    filename : str
        Name of file to get split positions

    num_sections : int
        The number of sections the file should be split into

    Returns
    -------
    list[tuple(int, int)]
        A list of start and end bytes which define each section

    """
    with open(filename, "r") as fptr:
        # Get size of file in bytes
        file_size = fileSize(fptr)

        # Divide size by num_sections to get aprox section sizes
        section_size = file_size // num_sections

        # Iterate through sizes and store pairs of start, end
        pairs = []
        start_byte = 0
        for i in range(num_sections):
            # Get approx end byte
            approx_end = start_byte + section_size
            if approx_end > file_size:
                approx_end = file_size

            # Move fileptr to approx end and then to start of line
            fptr.seek(approx_end)

            # Readline to get to next full line
            junk = fptr.readline()

            # Now iterate until the next line starts with a diff seq than this
            this_pos = fptr.tell()
            this_line = fptr.readline()
            next_pos = fptr.tell()
            next_line = fptr.readline()
            while (this_line.split(',')[0] == next_line.split(',')[0]):
                # iterate
                this_pos = next_pos
                this_line = next_line
                next_pos = fptr.tell()
                next_line = fptr.readline()
                # Check end of file
                if this_pos == next_pos:
                    break

            # Now this_line is different than next line, so next pos is end
            end_byte = next_pos

            # Save this value as end_byte
            pairs.append((start_byte, end_byte))

            # Print for testing
            fptr.seek(start_byte)
            start_byte = end_byte
        return pairs


def simplifyStream(stream):
    """ Simplify a stream by merging same sequences

    Parameters
    ----------
    stream : generator of Amplicon or ConservedEndAmplicons
        An iterable of Amplicons or ConservedEndAmplicons to simplify

    """
    # Iterate through stream and merge conserved sequences
    prev_item = None
    for item in stream:
        if item is None:
            # End of stream
            if prev_item is not None:
                yield prev_item
            return
        elif prev_item is None:
            # Just set as previous
            prev_item = item
        elif item == prev_item:
            # Same as previous so merge
            prev_item += item
        else:
            # Different so yield + reset
            yield prev_item
            prev_item = item

    # Yield last tuple if not None
    if prev_item is not None:
        yield prev_item


def removeSingletons(stream):
    """ Remove singletons from a stream

    Parameters
    ----------
    stream : generator of Amplicon or ConservedEndAmplicons
        An iterable of Amplicons or ConservedEndAmplicons to simplify

    Yields
        An iterable with all singletons removed

    """
    # Iterate through stream and merge conserved sequences
    prev_item = None
    singleton = True
    for item in stream:
        if item is None:
            # End of stream
            if prev_item is not None:
                yield prev_item
            return
        elif prev_item is None:
            # Just set as previous
            prev_item = item
            singleton = True
        elif item == prev_item:
            # Same as previous so yield
            yield prev_item
            prev_item = item
            singleton = False
        else:
            # Different so yield + reset
            if not singleton:
                yield prev_item
            prev_item = item
            singleton = True

    # Yield last item if not singleton
    if (prev_item is not None) and (not singleton):
        yield prev_item


def mergeSortedStreams(stream0, stream1):
    """ Merge two streams, maintaining sorted order

    Parameters
    ----------
    stream0 : Generator
        Generator of Amplicons or ConservedEndAmplicons

    stream1 : Generator
        Generator of Amplicons or ConservedEndAmplicons

    Yields
    ------
    Amplicons or ConservedEndAmplicons
        An iterable of sorted amplicons or ConservedEndAmplicons

    """
    # Pad streams with Nones so they don't end
    stream0 = itertools.chain(stream0, itertools.repeat(None))
    stream1 = itertools.chain(stream1, itertools.repeat(None))

    # Iterate while both are not None
    next_s0 = next(stream0)
    next_s1 = next(stream1)
    while (next_s0 is not None) or (next_s1 is not None):
        # Split on ordering
        if (next_s0 is None) or (next_s1 is not None) and (next_s1 < next_s0):
            # next_s1 is smaller
            yield next_s1
            next_s1 = next(stream1)
        else:
            # next_s0 is smaller
            yield next_s0
            next_s0 = next(stream0)


def intersectSortedStreams(stream0, stream1):
    """ Merge two ConservedEndAmplicons streams

    Parameters
    ----------
    stream0 : Generator
        Generator of ConservedEndAmplicons

    stream1 : Generator
        Generator of ConservedEndAmplicons

    Yields
    ------
    ConservedEndAmplicons
        An iterable of ConservedEndAmplicons

    """

    # Make sure each stream is simplified, i.e. duplicate elements are merged
    stream0 = simplifyStream(stream0)
    stream1 = simplifyStream(stream1)

    # Merge streams together
    merged = mergeSortedStreams(stream0, stream1)

    # Remove singletons and yield
    yield from simplifyStream(removeSingletons(merged))


def ampliconStream(filename, start=None, end=None):
    """ Converts a kmer file into a stream of Amplicons

    Parameters
    ----------
    filename : str
        Name of kmer file to read

    start : int, optional
        Starting byte of file to read

    end : int, optional
        End byte of file to read

    Yields
    ------
    Amplicon
        A stream of amplicons derived from the file with all duplicates merged

    """
    # Create a helper function to read from file
    def helper(filename, start, end):
        # Get tag from filename
        tag = simplename(filename)

        # Open file for reading
        with open(filename) as fptr:
            # Get start and end bytes
            if start is None:
                start = 0
            if end is None:
                end = fileSize(fptr)

            # Move file pointer to start
            fptr.seek(start)
            # Iterate until end is reached
            while fptr.tell() < end:
                # Get next line or break
                line = fptr.readline()
                if not line:
                    break
                # Convert line into an amplicon
                yield Amplicon.read(line, tag)

    # First get the raw amplicon stream
    stream = helper(filename, start, end)

    # Now simplify stream and yield
    yield from simplifyStream(stream)


def writeAmpliconStream(stream, output, mode="w"):
    """ Write an Amplicon stream to a file

    Parameters
    ----------
    stream : Iterable of amplicons
        An iterable of Amplicons to write

    output : str
        An output filename to write stream to

    """
    # Write all amplicons to output
    with open(output, mode) as fout:
        for amplicon in stream:
            amplicon.write(fout)


def writeAlignmentStream(stream, output, mode="w"):
    """ Write an alignment stream to a file

    Parameters
    ----------
    stream : Iterable of alignments
        An iterable of ConservedEndAmplicons to write

    output : str
        An output filename to write stream to

    """
    # Get amplicon stream
    def helper(align_stream):
        for align in align_stream:
            for ampl in align.amplicons:
                yield ampl

    ampl_stream = helper(stream)
    writeAmpliconStream(ampl_stream, output, mode=mode)



def alignmentStream(kmerfile, start=None, end=None, ingroup=None):
    """ Read a kmer file from start to end and yield alignments

    Parameters
    ----------
    filename : str
        Name of kmer file to read

    start : int, optional
        Starting byte of file to read

    end : int, optional
        End byte of file to read

    Yields
    ------
    Amplicon
        A stream of alignments derived from the file

    """
    # Read the kmerfile from start to end
    alignment = ConservedEndAmplicons(ingroup)
    for amplicon in ampliconStream(kmerfile, start, end):
        # Check if this amplicon can't fit into the alignment
        if not alignment.canAdd(amplicon):
            # Different so yield this alignment and reset
            yield alignment
            alignment = ConservedEndAmplicons(ingroup)
        # Add to alignment
        alignment += amplicon

    # Print last alignment if not empty
    if len(alignment):
        yield alignment





from krisp_vcf.find_diag_var import *
from krisp_vcf.find_diag_var import _parse_group_data, _check_variant, GroupedVariant
from krisp_vcf.find_diag_region import GroupedRegion, find_diag_region
import pysam
from itertools import islice
import unittest
from Bio import SeqIO


class TestGroupedRegion(unittest.TestCase):
    def setUp(self):
        pysam.set_verbosity(0)
        self.vcf = pysam.VariantFile('test_data/unfilt_allscafs_n666.vcf.gz')
        self.groups = _parse_group_data('test_data/test_metadata.tsv')
        self.groups = {g: v for g, v in self.groups.items() if g in ["NA1", "NA2", "EU1", "EU2"]}
        self.ref = 'test_data/PR-102_v3.1.fasta'
        self.variant1 = next(self.vcf)
        self.variant2 = next(self.vcf)
        self.variant3 = next(self.vcf)
        self.diag_var = next(islice(self.vcf, 1020, 1021))
        self.vcf_subset = islice(self.vcf, 10, 15)

    def test_init(self):
        x = GroupedRegion(GroupedVariant.from_vcf(self.vcf_subset, groups=self.groups), group="NA1", reference=self.ref)
        print(x.region_length())

    # def test_sliding_window(self):
    #     windower = GroupedRegion.sliding_window(GroupedVariant.from_vcf(self.vcf, groups=self.groups),
    #                                             groups=self.groups.keys(),
    #                                             reference=self.ref,
    #                                             span=30)
    #     for x in windower:
    #         print(x.group, x.region_length())
    #         print(len(x.variants))

    def test_find_diag_region(self):
        find_diag_region(self.vcf,
                         groups=self.groups,
                         reference=self.ref)




import shutil
import math
from collections import defaultdict
from prettytable import PrettyTable

class Annotation:

    def __init__(self, name, seq, start):
        self.name = name
        self.seq = seq
        self.start = start


def cumulative(lists):
    """
    https://www.geeksforgeeks.org/python-program-to-find-cumulative-sum-of-a-list/
    """
    length = len(lists)
    cu_list = [sum(lists[0:x:1]) for x in range(0, length + 1)]
    return cu_list[1:]


def _mask_same(seqs, ref):
    """If loci are the same, letter will be replaced with a period

    Parameters
    ----------
    seqs : dict of list of str
        sequences to align, named by group
    ref : list of str
        the reference used to call the variants

    Returns
    -------
    list of str
        Modified version of seqs with equivalent elements replaced with "."

    """
    for group in seqs.keys():
        for row_index in range(len(seqs[group])):
            if seqs[group][row_index] == ref[row_index]:
                middle = len(ref[row_index])
                seqs[group][row_index] = "." * middle

    return seqs


def pos_to_chunk_index(pos, ref):
    pos_to_chunk_key = {p - 1: i for i, p in enumerate(cumulative([len(c.strip()) for c in ref]))}
    for p, i in pos_to_chunk_key.items():
        if p >= pos:
            return {'chunk': i, 'offset': abs(p - pos)}


def _pad_sequences(seqs, ref, annots):
    """Adds padding for all sequences, with spacing or dashes
    Parameters
    ----------
    seqs : dict of list of str
        sequences to align, named by group
    ref : list of str
        list of alleles in the reference genome

    Return
    ------
    list of str
        Modified ref where indels are labelled "-", up to max. column length

    dict of list of str
        Modified seqs where indels are labelled "-", up to max. column length,
        any ambiguous or heterozygous loci are padded with " " to max. length

     """

    def _pad_ref_and_seqs(col_i, width, pad_str):
        ref[col_i] = ref[col_i].center(width, pad_str)
        for seq_name in seqs.keys():
            if seqs[seq_name][col_i] == " ": #used for printing oligos sequences
                seqs[seq_name][col_i] = seqs[seq_name][col_i].center(width, " ")
            else:
                seqs[seq_name][col_i] = seqs[seq_name][col_i].center(width,
                                                                     pad_str)
        return None

    # Pad reference and sequences to equal column widths
    for col_index in range(len(ref)):
        column = [ref[col_index]] + [seq[col_index] for seq in seqs.values()]
        max_width = max([len(x) for x in column])
        if any([">" in x or "/" in x for x in column]):
            _pad_ref_and_seqs(col_index, max_width, pad_str=" ")
        else:
            _pad_ref_and_seqs(col_index, max_width, pad_str="-")

    # Split annotation seq names into columns for printing later
    col_widths = [len(c) for c in ref]
    annot_out = [" " * len(c) for c in ref]
    for annot in annots:
        start = pos_to_chunk_index(annot.start, ref)
        end = pos_to_chunk_index(annot.start + len(annot.seq) - 1, ref)
        annot_col_widths = [col_widths[i] for i in range(start['chunk'], end['chunk'] + 1)]
        print_len = sum(annot_col_widths)
        name = ' ' + annot.name + ' '
        text = '└' + name.center(print_len - 2, '─') + '┘'
        text_inter = iter(text)
        text_cols = ["".join(next(text_inter) for i in range(w)) for w in annot_col_widths]
        for text_index, ref_index in enumerate(range(start['chunk'], end['chunk'] + 1)):
            annot_out[ref_index] = text_cols[text_index]

    return seqs, ref, annot_out


def _print_align(seqs, ref, annot_text, groups, ref_name="Reference"):
    """Adjusts labels, prints reference and sequences
    """


    def _print_one_line(seqs, ref, groups, ref_name="Reference"):
        """Repeated function to print the output format"""
        # Get counts of each group to print
        group_counts = {g: str(len(v)) for g, v in groups.items()}
        max_count_len = max([len(n) for n in group_counts.values()])

        # Get max length of labels
        labels = {k: f'{k} ({group_counts[k]})' if k in group_counts else f'{k}' for k in seqs.keys()}
        max_len = max([len(label) for label in list(labels.values()) + [ref_name]])

        # Print alignment
        output = []
        ref_name = ref_name.rjust(max_len, " ")
        #import pdb; pdb.set_trace()


        output.append(f"{ref_name}: " + ''.join(ref))
        for group_name, seq in seqs.items():
            group_text = labels[group_name].rjust(max_len, " ")
            output.append(f"{group_text}: " + ''.join(seq))

        # Print annotation names
        output.append(' ' * (max_len + 2) + ''.join(annot_text))
        # ref_len = sum([len(x) for x in ref])
        # annot_line = ([" "] * ref_len)
        # for annot in annots:
        #     size =
        #     for index, nucleotide in enumerate(annot.seq):
        #         annot_line[annot.start + index] = nucleotide

        return output

    def split(x, f):
        """
        https://stackoverflow.com/questions/19597097/python-equivalent-of-r-split-function
        """
        res = defaultdict(list)
        for v, k in zip(x, f):
            res[k].append(v)
        return res
#   align_width: consistent label length so sequences line up
#   col_widths: find length of columns in ref
#   row_index: length of character divided by aligned width, rounded down
#   chunked_ref: sort row_index into the assigned rounded groups
#   chunked_seqs: same as chunked_ref, but for dict of list (dict comprehension)
#   for loop makes the columns aesthetic
    term_width = shutil.get_terminal_size().columns
    labels = list(seqs.keys()) + [ref_name]
    label_width = max([len(seq) for seq in labels])
    align_width = term_width - label_width - 5
    col_widths = [len(x) for x in ref]
    row_index = [math.floor(x/align_width) for x in cumulative(col_widths)]
    chunked_ref = split(ref, row_index)
    chunked_seqs = {k: split(v, row_index) for k, v in seqs.items()}
    output = []
    for index in range(len(chunked_ref)):
        row_seqs = {k: v[index] for k, v in chunked_seqs.items()}
        output.extend(_print_one_line(row_seqs, chunked_ref[index], groups, ref_name=ref_name))
    return output


def format_seq_annot(annots, ref):
    """Go through annotated sequences and overwrite to add primer data

    Parameters:
    -----------
    annots: list of Annotation
        primers identified by Primer3/ package
    ref: list of str
        list of alleles in the reference genome
    Return:
         list(?)
    """
    ref_len = sum([len(x) for x in ref])
    output = ([" "] * ref_len)
    for annot in annots:
        start = pos_to_chunk_index(annot.start, ref)
        for index, nucleotide in enumerate(annot.seq):
            output[start['chunk'] + index] = nucleotide
    return output


def _render_primer3_stats(p3):
    left_stats = {k[14:]: v for k, v in p3.items() if 'PRIMER_LEFT_0_' in k}
    right_stats = {k[15:]: v for k, v in p3.items() if 'PRIMER_RIGHT_0_' in k}
    pair_stats = {k[14:]: v for k, v in p3.items() if 'PRIMER_PAIR_0_' in k}

    def render_col_names(names):
        return [x.title().replace('_', ' ') for x in names]

    def render_col_values(names):
        return [str(round(x, 5)) if isinstance(x, float) else x for x in names]

    primer_table = PrettyTable(['Direction'] + render_col_names(left_stats.keys()))
    primer_table.add_row(['Forward'] + render_col_values(left_stats.values()))
    primer_table.add_row(['Reverse'] + render_col_values(right_stats.values()))
    primer_table.align = 'l'

    pair_table = PrettyTable(render_col_names(pair_stats.keys()))
    pair_table.add_row(render_col_values(pair_stats.values()))
    pair_table.align = 'l'

    output = '\nPrimer statistics:\n' + \
             primer_table.get_string(border=False) + \
             '\n\nPair statistics:\n' + \
             pair_table.get_string(border=False)
    return output


def render_variant(seqs, ref, p3, groups, annots=None):
    """Displaying diagnostic variant in human-readable form

    Parameters
    ----------
    seqs : dict of list of str
        sequences to align, named by group
    ref : list of str
        the reference used to call the variants
    p3 : dict
        raw primer3 output
    annots : list of Annotation
        sequences to annotate alignment, keys are sequences, values are start
        position in the alignment

    Returns
    -------
    list of str
        one str for each line to print
    """

    # Make alignment
    seqs = _mask_same(seqs, ref)
    if annots is not None:
        seqs["oligos"] = format_seq_annot(annots, ref)
    seqs, ref, annot_text = _pad_sequences(seqs, ref, annots)
    output = _print_align(seqs, ref, annot_text, groups)

    # Make Primer3 output
    output += [_render_primer3_stats(p3)]

    return output


if __name__ == "__main__":
    ref = ["A", "C", "GTC", "T", "T", "G", "G", "C", "C", "A", "C", "GTC", "T",
           "T", "G", "G", "C", "C", "A", "C", "GTC", "T", "T", "G", "G", "C",
           "C", "A", "C", "GTC", "T", "T", "G", "G", "C", "C", "A", "C", "GTC",
           "T", "T", "G", "G", "C", "C", "A", "C", "GTC", "T", "T", "G", "G",
           "C", "C", "A", "C", "GTC", "T", "T", "G", "G", "C", "C"]
    seqs = {"seq 1": ["A", "G", "GTC", "T/G", "T", "GC", "G", "A", "C", "A",
                      "G", "GTC", "T/G", "T", "GC", "G", "A", "C", "A", "G",
                      "GTC", "T/G", "T", "GC", "G", "A", "C", "A", "G", "GTC",
                      "T/G", "T", "GC", "G", "A", "C", "A", "G", "GTC", "T/G",
                      "T", "GC", "G", "A", "C", "A", "G", "GTC", "T/G", "T",
                      "GC", "G", "A", "C", "A", "G", "GTC", "T/G", "T", "GC",
                      "G", "A", "C"],
            "seq 2": ["A", "G", "G", "T", "", "G", "<123G12?>", "A", "C", "A",
                      "G", "G", "T", "", "G", "<123G12?>", "A", "C", "A", "G",
                      "G", "T", "", "G", "<123G12?>", "A", "C", "A", "G", "G",
                      "T", "", "G", "<123G12?>", "A", "C", "A", "G", "G", "T",
                      "", "G", "<123G12?>", "A", "C", "A", "G", "G", "T", "",
                      "G", "<123G12?>", "A", "C", "A", "G", "G", "T", "", "G",
                      "<123G12?>", "A", "C"]}
    annotation = [Annotation("Left primer", "ATTGCATGA", 33),
                  Annotation("crRNA", "TGGTCCATGAT", 45),
                  Annotation("Right primer", "ATTGTAAC", 12)]
    render_variant(seqs, ref, annots=annotation)





import pdb
import sys
import re
import primer3
from statistics import mean
from Bio.Data import IUPACData
from prettytable import PrettyTable
from pudb.remote import set_trace

UNKNOWN_CHAR = "?"
iupac_key = {tuple((x for x in sorted(v))): k for k, v in
             IUPACData.ambiguous_dna_values.items()}
iupac_key[(UNKNOWN_CHAR,)] = 'N'

primer3_col_names = [
    'PRIMER_PAIR_0_PRODUCT_SIZE',
    'PRIMER_PAIR_0_PENALTY',
    'PRIMER_LEFT_0_SEQUENCE', 'PRIMER_RIGHT_0_SEQUENCE',
    'PRIMER_LEFT_0_PENALTY', 'PRIMER_RIGHT_0_PENALTY',
    'PRIMER_LEFT_0_TM', 'PRIMER_RIGHT_0_TM',
    'PRIMER_LEFT_0_GC_PERCENT', 'PRIMER_RIGHT_0_GC_PERCENT',
    'PRIMER_LEFT_0_SELF_ANY_TH', 'PRIMER_RIGHT_0_SELF_ANY_TH',
    'PRIMER_LEFT_0_SELF_END_TH', 'PRIMER_RIGHT_0_SELF_END_TH',
    'PRIMER_LEFT_0_HAIRPIN_TH', 'PRIMER_RIGHT_0_HAIRPIN_TH',
    'PRIMER_LEFT_0_END_STABILITY', 'PRIMER_RIGHT_0_END_STABILITY',
    'PRIMER_PAIR_0_COMPL_ANY_TH', 'PRIMER_PAIR_0_COMPL_END_TH',
]
primer3_col_key = {n: n.replace("PRIMER_", "").replace("_0", "").lower() for n in primer3_col_names}

class ForkedPdb(pdb.Pdb):
    """A Pdb subclass that may be used
    from a forked multiprocessing child

    """
    def interaction(self, *args, **kwargs):
        _stdin = sys.stdin
        try:
            sys.stdin = open('/dev/stdin')
            pdb.Pdb.interaction(self, *args, **kwargs)
        finally:
            sys.stdin = _stdin

def collapse_to_iupac(seqs):
    """Combine sequences into a consensus using IUPAC ambiguity codes

    Parameters:
    -----------
    seqs : list of str
        The sequences to combine.

    Returns
    -------
    str
        The combined sequence
    """
    seq_lens = [len(x) for x in seqs]
    max_len = max(seq_lens)
    if len(set(seq_lens)) != 1:  # TODO: replace with alignment
        return '-' * max_len
    output = []
    for i in range(max_len):
        column = {s[i] for s in seqs}
        if "*" in column or "N" in column or UNKNOWN_CHAR in column:
            output.append('N')
        else:
            output.append(iupac_key[tuple(sorted(column))])
    return "".join(output)


def parse_primer3_settings(file_path):
    """
    Reads primer3 BoulderIO format, assuming only global settings (starts with PRIMER_),
    and returns a dict in the format used by primer3-py.
    """
    def to_number_if_can(x):
        try:
            if int(float(x)) == float(x) and '.' not in x:
                return int(x)
            else:
                return float(x)
        except ValueError:
            return x

    with open(file_path) as handle:
        options = dict([tuple(l.strip().split('=')) for l in handle.readlines()])
        for opt, val in options.items():
            if ' ' in val or ';' in val:
                val = re.split('[ ;]+', val)
                val = [to_number_if_can(v) for v in val]
                if ',' in val or '-' in val[0]:
                    val = [[to_number_if_can(x) for x in re.split('[,-]+', v)] for v in val]
            elif ',' in val or '-' in val:
                val = re.split('[,\-]+', val)
                val = [to_number_if_can(v) for v in val]
            else:
                val = to_number_if_can(val)
            options[opt] = val
    return options

def _format_p3_output(p3_out):
    """Reformat data for best primer pair for CSV output"""
    return {primer3_col_key[n]: p3_out[n] for n in primer3_col_names}

def run_primer3(template, target_start, target_len,
                options=None,
                tm=(53, 68),
                gc=(40, 70),
                amp_size=(80, 300),
                primer_size=(25, 35),
                max_sec_tm=40,
                gc_clamp=1,
                max_end_gc=4):
    if options is None:
        global_options = {
            'PRIMER_TASK': 'generic',
            'PRIMER_PICK_LEFT_PRIMER': 1,  # 1 == True
            # 'PRIMER_PICK_INTERNAL_OLIGO': 1,  # 1 == True
            'PRIMER_PICK_RIGHT_PRIMER': 1,  # 1 == True
            'PRIMER_LIBERAL_BASE': 1,  # 1 == True
            'PRIMER_OPT_SIZE': mean(primer_size),
            'PRIMER_MIN_SIZE': primer_size[0],
            'PRIMER_MAX_SIZE': primer_size[1],
            # 'PRIMER_INTERNAL_MAX_SIZE': len(crrna_seq),
            'PRIMER_OPT_TM': mean(tm),
            'PRIMER_MIN_TM': tm[0],
            'PRIMER_MAX_TM': tm[1],
            'PRIMER_MIN_GC': gc[0],
            'PRIMER_MAX_GC': gc[1],
            'PRIMER_MAX_POLY_X': 4,  # The maximum allowable length of a mononucleotide repeat
            'PRIMER_MAX_NS_ACCEPTED': 0,  # The maximum number of Ns
            'PRIMER_THERMODYNAMIC_OLIGO_ALIGNMENT': 1,  # 1 == True
            'PRIMER_MAX_SELF_ANY_TH': max_sec_tm,  # describes the tendency of a primer to bind to itself
            'PRIMER_MAX_SELF_END_TH': max_sec_tm,  # 3' binding to itself (primer dimers)
            'PRIMER_PAIR_MAX_COMPL_ANY_TH': max_sec_tm,  # the tendency of the left primer to bind to the right primer
            'PRIMER_PAIR_MAX_COMPL_END_TH': max_sec_tm,  # primer heterodimers
            'PRIMER_MAX_HAIRPIN_TH': max_sec_tm,  # hairpins
            'PRIMER_PRODUCT_SIZE_RANGE': [amp_size],
            'PRIMER_GC_CLAMP': gc_clamp,
            'PRIMER_MAX_END_GC': max_end_gc,
        }
    else:
        global_options = parse_primer3_settings(options)

    p3_output = primer3.bindings.design_primers(
        {
            'SEQUENCE_TEMPLATE': "".join(template),
            # 'SEQUENCE_INTERNAL_OLIGO': "".join(crrna_seq),
            'SEQUENCE_TARGET': [target_start, target_len]
        },
        global_options
    )
    return p3_output


class Amplicon:
    def __init__(self, primer, diagnostic, reverse, *labels):
        """ Store values """
        # Store sequences
        self.primer = primer
        self.diagnostic = diagnostic
        self.reverse = reverse

        # Store labels as a list of filenames
        self.labels = sorted(labels)

    @property
    def sequence(self):
        """ Return the sequence for this amplicon """
        return f"{self.primer}{self.diagnostic}{self.reverse}"

    def _labelsToString(self):
        """ Helper function to convert labels into a string """
        # Convert labels to a dictionary for counting
        counts = {}
        for label in self.labels:
            if label not in counts:
                counts[label] = 1
            else:
                counts[label] += 1

        # Merge all labels into a single label list
        label_list = []
        for filename, count in sorted(counts.items()):
            if count == 1:
                label_list.append(f"{filename}")
            else:
                label_list.append(f"{filename}({count})")
        return ';'.join(label_list)

    def _stringToLabels(string):
        """ Helper function to convert a string into a series of labels """
        # Convert string to a list of labels
        labels = []

        # Read through tags to get label
        for label in string.split(';'):
            # Strip whitespace
            label = label.strip()

            # Check for multiplier
            if '(' in label:
                name, multiplier = label.split('(')
                multiplier = int(multiplier.strip(')'))
                labels += [name] * multiplier
            else:
                labels += [label]
        return labels

    def __str__(self):
        """ Returns a string representation of this amplicon """
        return f"{self.sequence} : {self._labelsToString()}"

    def __eq__(self, other):
        """ Returns true if same sequence as other

        Parameters
        ----------
        other : Amplicon
            The other amplicon to compare with

        Returns
        -------
        bool
            True if self and other have the same sequence, else False

        Raises
        ------
        TypeError
            If other is not an Amplicon instance

        """
        # Check Valid type
        if not isinstance(other, Amplicon):
            raise TypeError(f"Can't compare Amplicon with type {type(other)}")
        return self.sequence == other.sequence

    def __add__(self, other):
        """ Merges two amplicons and their labels

        Parameters
        ----------
        other : Amplicon
            The other amplicon to add

        Returns
        -------
        Amplicon
            A new Amplicon instance containing the merged labels

        Raises
        ------
        TypeError
            If other is not an Amplicon instance
        ValueError
            If the amplicons have different sequences

        """
        # Check Valid type
        if not isinstance(other, Amplicon):
            raise TypeError(f"Can't compare Amplicon with type {type(other)}")

        # Check valid addition
        if not self.__eq__(other):
            raise ValueError("Can't merge Amplicons with different sequences")

        # Create a new amplicon to return
        return Amplicon(self.primer,
                        self.diagnostic,
                        self.reverse,
                        *self.labels,
                        *other.labels)

    def __lt__(self, other):
        """ Returns true if self comes before other in terms of primer ordering

        Parameters
        ----------
        other : Amplicon
            The other amplicon to compare with

        Returns
        -------
        bool
            True if self comes before other in terms of primer sequences

        Raises
        ------
        TypeError
            If other is not an Amplicon instance

        """
        # Check Valid type
        if not isinstance(other, Amplicon):
            raise TypeError(f"Can't compare Amplicon with type {type(other)}")

        # Check primers
        return (self.primer, self.reverse) < (other.primer, other.reverse)

    def read(string, filename):
        """ Class method to read a string from a file to an Amplicon

        Parameters
        ----------
        string : str
            An input string to create an amplicon. Assumes the format:
            primer, diagnostic, reverse, tag0; tag1; tag2

        filename : str
            The default label to add to the Amplicon

        """
        # Split string by commas
        fields = string.strip().split(',')
        if len(fields) in [3, 4]:
            # Split fields
            primer, diag, reverse, *tags = fields

            # Format tags
            labels = []
            if len(tags) == 0:
                # Set label to be filename
                labels = [filename]
            else:
                labels = Amplicon._stringToLabels(tags[0])
            # Create amplicon and return
            return Amplicon(primer, diag, reverse, *labels)
        else:
            raise ValueError(f"Unrecognised string format : {string}")
            #removed "return None" as it was unreachable

    def write(self, fileptr):
        """ Function to write self to a file such that it can be read later

        Parameters
        ----------
        fileptr : file object
            A file object to write this instance to

        """
        # Build fields to write
        # First set the sequence
        fields = [self.primer, self.diagnostic, self.reverse]

        # Now add the tags as ';' separated values
        if len(self.labels) != 0:
            fields.append(self._labelsToString())

        # Finally, write to file
        print(*fields, sep=',', file=fileptr)


class ConservedEndAmplicons:
    """
    Helper class to print sequences which have conserved ends and are all of
    equal length. Sequences are added using the add method which takes as
    arguments the forward primer, the diagnostic region, the reverse primer,
    and a label (typically a filename). Alignments can then be printed by
    calling the str() method or printing directly.
    """
    # Class level variables
    ENABLE_DOT = False
    P3_ARGS = {}

    def __init__(self, ingroup=None):
        """ Store data as a list of amplicons """
        self.amplicons = []
        self.ingroup = None
        if ingroup is not None:
            self.ingroup = frozenset(ingroup)
        self.p3 = None

    def labels(self):
        """ Returns the set of labels contained in this alignment """
        labels = set()
        for ampl in self.amplicons:
            labels |= set(ampl.labels)
        return labels

    def primerPair(self):
        """ Return the forward and reverse primers as a tuple """
        if len(self.amplicons) > 0:
            forward = self.amplicons[0].primer
            reverse = self.amplicons[0].reverse
            return (forward, reverse)
        else:
            raise ValueError("No amplicons added yet")

    def primerLength(self):
        """ Return the length of the primer region """
        if len(self.amplicons) > 0:
            return len(self.amplicons[0].primer)
        else:
            raise ValueError("No amplicons added yet")

    def diagnosticLength(self):
        """ Return the length of the diagnostic region """
        if len(self.amplicons) > 0:
            return len(self.amplicons[0].diagnostic)
        else:
            raise ValueError("No amplicons added yet")

    def ampliconLength(self):
        """ Return the length of the entire amplicon """
        if len(self.amplicons) > 0:
            return len(self.amplicons[0].sequence)
        else:
            raise ValueError("No amplicons added yet")

    def canAdd(self, other):
        """ Returns true if other can be added to this

        Parameters
        ----------
        other : Amplicon or ConservedEndAmplicons
            The instance to test addition

        Returns
        -------
        bool
            True if it can be added, else False

        Raises
        ------
        TypeError
            If input is not an Amplicon or ConservedEndAmplicons instance


        """
        # Split on type
        if isinstance(other, Amplicon):
            # Check if empty
            if len(self.amplicons) == 0:
                return True

            # Check if primers are conserved
            primer = other.primer
            reverse = other.reverse
            return (primer, reverse) == self.primerPair()
        elif isinstance(other, ConservedEndAmplicons):
            # Check if empty
            if len(self.amplicons) == 0:
                return True

            # Check if primers are conserved
            return self.primerPair() == other.primerPair()
        else:
            raise TypeError("Unexpected type input")

    def add(self, other):
        """ Add other to this instance

        Parameters
        ----------
        other : Amplicon or ConservedEndAmplicons
            The instance to add

        Returns
        -------
        None

        Raises
        ------
        TypeError
            If input is not an Amplicon or ConservedEndAmplicons instance

        """
        # Split on input type
        if isinstance(other, Amplicon):
            # Check if this can be merged with other amplicons
            for i in range(len(self.amplicons)):
                if other == self.amplicons[i]:
                    self.amplicons[i] += other
                    break
            else:
                # Can't merge so just add to list
                self.amplicons.append(other)
        elif isinstance(other, ConservedEndAmplicons):
            # Add all amplicons to self
            for amplicon in other.amplicons:
                self.add(amplicon)
        else:
            raise TypeError("Unexpected input type")

    def diagnosticColumns(self):
        """ Returns a list of diagnostic positions """
        # Get the diagnostic sequence for each amplicon
        diagnostics = [ampl.diagnostic for ampl in self.amplicons]

        # Determine diagnostic positions
        diags = []
        for i, bases in enumerate(zip(*diagnostics)):
            if len(set(bases)) > 1:
                diags.append(i)
        return diags

    def ingroupUniqueColumns(self):
        """ Returns a list of columns unique to the ingroup """
        # Return empty list if ingroup is not set
        if self.ingroup is None:
            return []

        # Get the diagnostic sequence for each amplicon
        ingroup_diag = []
        outgroup_diag = []
        for amplicon in self.amplicons:
            # Iterate through each label
            for label in amplicon.labels:
                if label in self.ingroup:
                    ingroup_diag.append(amplicon.diagnostic)
                else:
                    outgroup_diag.append(amplicon.diagnostic)

        # Iterate through all positions
        diags = []
        for i in range(self.diagnosticLength()):
            # Get bases for ingroup and outgroup
            ingroup_bases = set([d[i] for d in ingroup_diag])
            outgroup_bases = set([d[i] for d in outgroup_diag])
            # Check if conserved in ingroup and not in outgroup
            if ingroup_bases.isdisjoint(outgroup_bases):
                diags.append(i)
        return diags

    def makeBracket(self):
        """ Make graphical bracket to output with alignment """
        # Get start and end positions of diagnostic sequences
        start = self.primerLength()
        end = start + self.diagnosticLength()

        # Create an empty bracket
        bracket = ' ' * (start-1) + '{' + '-' * (end-start) + '}'

        # Add * at diagnostic positions
        bracket = list(bracket)
        for d in self.diagnosticColumns():
            bracket[start + d] = "*"

        # Add # at unique positions in ingroup
        for d in self.ingroupUniqueColumns():
            bracket[start + d] = "#"
        return ''.join(bracket)

    def setIngroup(self, grouping):
        """ Set the 'ingroup' for the alignment, only changes str format """
        if grouping is not None:
            self.ingroup = frozenset(grouping)

    def ingroup_consensus(self):
        return self.consensus(self.ingroup)

    def consensus(self, labels=None):
        if labels is None:
            amps = self.amplicons
        else:
            amps = [x for x in self.amplicons if set(x.labels).issubset(labels)]
        forward_seq = collapse_to_iupac([x.primer for x in amps])
        diag_seq = collapse_to_iupac([x.diagnostic for x in amps])
        reverse_seq = collapse_to_iupac([x.reverse for x in amps])
        return {'forward' : forward_seq, 'diagnostic' : diag_seq, 'reverse' : reverse_seq}

    def find_primers(self):
        template = "".join(self.ingroup_consensus().values())
        self.p3 = run_primer3(template, target_start=self.primerLength(),
                              target_len=self.diagnosticLength(), **ConservedEndAmplicons.P3_ARGS)
        return self.p3['PRIMER_PAIR_NUM_RETURNED'] != 0

    def _render_primer3_stats(self):
        if self.p3 is None:
            raise ValueError('Primer3 not run before results are to be rendered.')

        left_stats = {k[14:]: v for k, v in self.p3.items() if 'PRIMER_LEFT_0_' in k}
        right_stats = {k[15:]: v for k, v in self.p3.items() if 'PRIMER_RIGHT_0_' in k}
        pair_stats = {k[14:]: v for k, v in self.p3.items() if 'PRIMER_PAIR_0_' in k}

        def render_col_names(names):
            return [x.title().replace('_', ' ') for x in names]

        def render_col_values(names):
            return [str(round(x, 5)) if isinstance(x, float) else x for x in names]

        primer_table = PrettyTable(['Direction'] + render_col_names(left_stats.keys()))
        primer_table.add_row(['Forward'] + render_col_values(left_stats.values()))
        primer_table.add_row(['Reverse'] + render_col_values(right_stats.values()))
        primer_table.align = 'l'

        pair_table = PrettyTable(render_col_names(pair_stats.keys()))
        pair_table.add_row(render_col_values(pair_stats.values()))
        pair_table.align = 'l'

        #set_trace(term_size=(80, 60))

        output = '\nPrimer statistics:\n' + \
                 primer_table.get_string(border=False) + \
                 '\n\nPair statistics:\n' + \
                 pair_table.get_string(border=False)
        return output


    def render_alignment(self):
        """ Return a string representation of an alignment """
        # Try splitting amplicons based on ingroup, outgroup
        result = []
        if self.ingroup is not None:
            in_result = []
            out_result = []
            for ampl in sorted(self.amplicons, key=lambda x: x.labels):
                if set(ampl.labels) & set(self.ingroup):
                    in_result.append(str(ampl))
                else:
                    out_result.append(str(ampl))
            result = in_result + out_result
        else:
            for ampl in sorted(self.amplicons, key=lambda x: x.labels):
                result.append(str(ampl))

        # Switch to dot alignment if enabled
        if ConservedEndAmplicons.ENABLE_DOT:
            # Mark conserved nucleotides as '.'
            top_seq = result[0]
            new_result = [top_seq]
            for seq in result[1:]:
                seq = list(seq)
                for i in range(self.ampliconLength()):
                    b0 = top_seq[i]
                    b1 = seq[i]
                    if b0 == b1:
                        seq[i] = '.'
                new_result.append(''.join(seq))
            result = new_result
        else:
            # Add diagnostic bracket to result
            result.append(self.makeBracket())

        # Add primer 3 primer annotations
        if self.p3 is not None:
            forward_seq = self.p3['PRIMER_LEFT_0_SEQUENCE']
            reverse_seq = self.p3['PRIMER_RIGHT_0_SEQUENCE']
            forward_start = self.p3['PRIMER_LEFT_0'][0]
            reverse_start = self.p3['PRIMER_RIGHT_0'][0] - self.p3['PRIMER_RIGHT_0'][1]
            forward_annot = '└' + 'Forward'.center(len(forward_seq) - 2, '─') + '┘'
            reverse_annot = '└' + 'Reverse'.center(len(reverse_seq) - 2, '─') + '┘'
            text_out = ' ' * (forward_start) + \
                       forward_annot + \
                       ' ' * (reverse_start - forward_start - len(forward_seq) + 1) + \
                       reverse_annot
            if ConservedEndAmplicons.ENABLE_DOT:
                result.append(text_out)
            else:
                result[-1] = result[-1].ljust(len(text_out))
                result[-1] = "".join([annot if bracket == ' ' else bracket
                                      for bracket, annot in zip(result[-1], text_out)])


        # Add primer3 statistics for primers
        if self.p3 is not None:
            result.append(self._render_primer3_stats())

        # Add newline to separate inputs
        result[-1] += '\n'

        # Join result with newlines and return
        return '\n'.join(result)

    def render_csv(self, sep=','):
        if len(self.amplicons) == 1:
            output = list(self.consensus().values())
        else:
            output = list(self.ingroup_consensus().values())
        if self.p3 is not None:
            output.extend(_format_p3_output(self.p3).values())
        output = [str(x) for x in output]
        return sep.join(output)

    def __str__(self):
        return self.render_alignment()

    def __len__(self):
        """ Return number of alignments """
        return len(self.amplicons)

    def __eq__(self, other):
        """ Return True if self and other have the same primer pairs """
        return self.primerPair() == other.primerPair()

    def __lt__(self, other):
        """ Return True if self comes before other in terms of primers """
        return self.primerPair() < other.primerPair()

    def __iadd__(self, other):
        """ Add other to self and return self """
        # Add other to self and return
        self.add(other)
        return self



from .shared import *


def filterAlignmentStream(stream, ingroup):
    """ Filter alignment stream based on ingroup

    Parameters
    ----------
    stream : generator of ConservedEndAmplicons
        A generator of alignments to filter
    ingroup : set
        A set-like object containing the simple names which denote the
        in-group, e.g. the set of genomes which should all contain a unique
        SNP which is identifiable

    Yields
    ------
    ConservedEndAlignments
        The filtered stream of alignments

    """
    # Iterate through every alignment in the stream
    for alignment in stream:
        # Set the ingroup
        alignment.setIngroup(ingroup)
        # Determine column numbers of unique SNP's : empty list -> None
        if len(alignment.ingroupUniqueColumns()) > 0:
            yield alignment


def filterAlignments(kmerfile, output, ingroup):
    # Read in alignments
    alignments = alignmentStream(kmerfile)

    # Filter alignment stream if ingroup is not empty
    if len(ingroup):
        alignments = filterAlignmentStream(alignments, ingroup)

    # Write to output
    writeAlignmentStream(alignments, output)



#! /bin/env python3
import sys
import argparse
import fileinput
import itertools
import multiprocessing
import subprocess


# Dictionary of Watson-Crick complements for DNA
COMP_MAP = {'A': 'T', 'T': 'A', 'a': 't', 't': 'a',
            'G': 'C', 'C': 'G', 'g': 'c', 'c': 'g',
            'R': 'Y', 'Y': 'R', 'r': 'y', 'y': 'r',
            'M': 'K', 'K': 'M', 'm': 'k', 'k': 'm',
            'S': 'S', 'W': 'W', 's': 's', 'w': 'w',
            'B': 'V', 'V': 'B', 'b': 'v', 'v': 'b',
            'D': 'H', 'H': 'D', 'd': 'h', 'h': 'd',
            'N': 'N', 'n': 'n'}

# IUPAC nucleobase generalizations
IUPAC_BASE = {'R': ['A', 'G'],
              'Y': ['C', 'T'],
              'S': ['G', 'C'],
              'W': ['A', 'T'],
              'K': ['G', 'T'],
              'M': ['A', 'C'],
              'B': ['C', 'G', 'T'],
              'D': ['A', 'G', 'T'],
              'H': ['A', 'C', 'T'],
              'V': ['A', 'C', 'G'],
              'N': ['A', 'C', 'G', 'T'],
              'r': ['a', 'g'],
              'y': ['c', 't'],
              's': ['g', 'c'],
              'w': ['a', 't'],
              'k': ['g', 't'],
              'm': ['a', 'c'],
              'b': ['c', 'g', 't'],
              'd': ['a', 'g', 't'],
              'h': ['a', 'c', 't'],
              'v': ['a', 'c', 'g'],
              'n': ['a', 'c', 'g', 't']}


def sortPipe(np=None, mem=None, cols=None):
    """ Create a sorting subprocess Pipe

    Parameters
    ----------
    np : int (optional)
        The number of cores to use for sorting

    mem : str (optional)
        The amount of memory to use during sorting. See linux sort '-S'

    cols : list of int (optional)
        Columns to sort on, assumes 0-indexing and ',' seperating

    Returns
    --------
    subprocess Pipe
        An active subprocess pipe that can be written to and read from

    """
    # Get default values
    command = f"LC_ALL=C sort"
    if mem is not None:
        command += f" -S {mem}"
    if np is not None:
        command += f" --parallel={np}"
    if (cols is not None):
        command += f" -t,"
        for c in cols:
            command += f" -k{c+1},{c+1}"
    # Start subprocess and return
    return subprocess.Popen(command,
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            text=True,
                            shell=True)


def sortInPlace(filename, np=None, mem=None, cols=None):
    """ Sorts a file in place

    Parameters
    ----------
    filename : str
        The file to sort

    np : int (optional)
        The number of cores to use for sorting

    mem : str (optional)
        The amount of memory to use during sorting. See linux sort '-S'

    cols : list of int (optional)
        Columns to sort on, assumes 0-indexing and ',' seperating

    Returns
    --------
    None
        The input file is sorted in place

    """

    # Get default values
    command = f"LC_ALL=C sort {filename} -o {filename}"
    if mem is not None:
        command += f" -S {mem}"
    if np is not None:
        command += f" --parallel={np}"
    if (cols is not None):
        command += f" -t,"
        for c in cols:
            command += f" -k{c+1},{c+1}"
    # Run subprocess
    process = subprocess.Popen(command, shell=True)
    process.wait()


class kstream:
    """
    A highly flexible class to read and parse kmers from a fasta file. The
    class operates on the idea of filtering and manipulating kmer streams. The
    input fasta file is converted to a kmer stream which is then manipulated
    based on the passed arguments.
    """

    def __init__(self, sequences=None, kmers=None, complements=False,
                 canonicals=False, allow=None, disallow=None, omitsoft=False,
                 mapsoft=False, expandiupac=False, split=None, sort=False,
                 sortmem=None, sortcols=None, sortnp=1, parallel=1):
        """
        Parameters
        ----------
        sequences : iterable of str, or fasta filename
            An iterable of sequence strings to parse or a fasta file

        kmers : int or [int]
            The desired kmer length or a list of desired kmer lengths

        complements : bool (default False)
            True will add the reveres complements to the resulting stream

        canonicals : bool (default False)
            True will convert all kmers to canonical form (alphabetical order)

        allow : str
            A list of allowed characters in kmers. Kmers will be omitted if
            they contain any character outside of allowed, e.g allow='ATGC'
            will only allow kmer with an 'A', 'T', 'G', 'C'.

        disallow : str
            A list of disallowed characters in kmers. The opposite of allow.

        omitsoft : bool (default False)
            Disallow any kmer containing a lowercase character.

        mapsoft : bool (default False)
            Map lowercase characters to uppercase in all kmers.

        expandiupac : bool (default False)
            Expand any IUPAC characters and yield each as a separate kmer.

        split : [int]
            A list of positions to split the resulting kmers on to create a
            series of ',' seperated sub-kmers.

        sort : bool (default False)
            Sort the resulting character stream. Requires linux sort utility.
            See sortmem, sortcols, sortnp for additional options.

        sortmem : str
            The amount of memory to delegate to the sorting routine. See sort
            linux command. Can be amount: "1G", or a %: "50%".

        sortcols : [int]
            A list of integer positions (0-indexed) to sort on. Default sorts
            on entire kmer. Useful with split command to sort on 1st, 3rd
            columns, e.g. sortcols=[0,2].

        sortnp : int
            Number of cores to dedicate to sorting

        parallel : int
            Number of cores to dedicate to kmer extraction

        Returns
        -------
        kstream instance
            Any kstream instance can be iterated through as though it was a
            list, assuming sequences was not None when initiated. Or, a kstream
            instance can be initiated and called repeatedly as a functor to
            retrieve an iterable instance.

        """

        # Parse stream options and add to parsers
        # Each parser is a function which takes an input stream, modifies it,
        # And yields the result. For example, self._complements will take a
        # Stream of kmers, and yield every input kmer, and its complement.
        self.parsers = []
        if kmers is not None:
            if isinstance(kmers, int):
                self.kmers = [kmers]
            else:
                self.kmers = list(kmers)
            self.parsers.append(self._kmers)
        if omitsoft is True:
            if mapsoft is True:
                raise ValueError("can't omit and map soft masked nucleotides")
            self.parsers.append(self._omitsoft)
        if mapsoft is True:
            self.parsers.append(self._mapsoft)
        if complements is True:
            if canonicals is True:
                raise ValueError("canonicals conflicts with complements")
            self.parsers.append(self._complements)
        if allow is not None:
            self.allow = set(allow)
            self.parsers.append(self._allow)
        if disallow is not None:
            self.disallow = set(disallow)
            self.parsers.append(self._disallow)
        if expandiupac is True:
            self.parsers.append(self._expandiupac)
        if canonicals is True:
            self.parsers.append(self._canonicals)
        if split is not None:
            if isinstance(split, int):
                self.split = [split]
            else:
                self.split = list(split)
            self.parsers.append(self._split)

        # Check if sorting is specified
        self.sort = sort
        if self.sort is True:
            self.sortnp = sortnp
            self.sortmem = sortmem
            self.sortcols = sortcols

        # Set parallel flag
        self.parallel = parallel

        # Store sequences for future call to iter
        self.sequences = sequences

    def write(self, filename, sequences=None):
        """ Write sequences to a file

        Parameters
        ----------
        filename : str
            Name of file to write kmers to

        sequences : [str] or str (Default None)
            The sequences to write, if None, assumes kstream was instanciated
            with sequences.

        Returns
        -------
        int
            The number of kmers written to the file

        """
        # Parse sequences
        if sequences is None:
            sequences = self.sequences
        sequences = self._parsed_kmers(sequences)

        # Check if RNA
        is_rna, sequences = self._detect_RNA(sequences)

        # Map to DNA if input is RNA
        if is_rna:
            sequences = self._to_DNA(sequences)

        # Keep count of sequences found
        count = 0

        # Open file for writing
        with open(filename, 'w') as fout:
            # Split on parallel
            if self.parallel == 1:
                # Apply filters
                for f in self.parsers:
                    sequences = f(sequences)

                # Map back to RNA
                if is_rna:
                    sequences = self._to_RNA(sequences)

                # Write sequences
                for seq in sequences:
                    print(f"{seq}", file=fout)
                    count += 1
            else:
                # Run parsing in parallel using a pool of workers
                with multiprocessing.Pool(self.parallel) as pool:
                    # Parse sequences using pools imap, result is nested list
                    for result in pool.imap_unordered(self._parallel_job,
                                                      sequences):
                        # Map back to RNA
                        if is_rna:
                            result = self._to_RNA(result)

                        # Write to file
                        for seq in result:
                            print(f"{seq}", file=fout)
                            count += 1
            # Make sure that file contents have been written
            fout.flush()

        # Check if sort
        if self.sort:
            # Sort file in place
            sortInPlace(filename,
                        np=self.sortnp,
                        mem=self.sortmem,
                        cols=self.sortcols)

        # Return number of kmers found
        return count

    def __call__(self, sequences):
        """ Convert sequences into a kmer stream

        Parameters
        ----------

        sequences : iterable
            The iterable of sequences or a fasta filename

        Yields
        ------
        str
            The resulting kmers

        """
        # Parse sequences
        sequences = self._parsed_kmers(sequences)

        # Check if RNA
        is_rna, sequences = self._detect_RNA(sequences)

        # Map to DNA if input is RNA
        if is_rna:
            sequences = self._to_DNA(sequences)

        # Split on parallel
        if self.parallel == 1:
            # Apply filters
            for f in self.parsers:
                sequences = f(sequences)

            # Check if sort
            if self.sort:
                # Create sort pipe
                sort_pipe = sortPipe(np=self.sortnp,
                                     mem=self.sortmem,
                                     cols=self.sortcols)
                # Write sequences to sort
                for seq in sequences:
                    print(f"{seq}", file=sort_pipe.stdin)
                sort_pipe.stdin.close()

                # set values back into sequences
                sequences = (line.strip() for line in sort_pipe.stdout)

            # Map back to RNA
            if is_rna:
                sequences = self._to_RNA(sequences)

            # Yield sequences
            yield from sequences
        else:
            # Run parsing in parallel using a pool of workers
            sort_pipe = None
            if self.sort:
                sort_pipe = sortPipe(np=self.sortnp,
                                     mem=self.sortmem,
                                     cols=self.sortcols)
            with multiprocessing.Pool(self.parallel) as pool:
                # Parse sequences using pools imap, results is a nested list
                for result in pool.imap_unordered(self._parallel_job,
                                                  sequences):

                    # Map back to RNA
                    if is_rna:
                        result = self._to_RNA(result)

                    # Check if sort
                    if self.sort:
                        # Write sequences to sort
                        for seq in result:
                            print(f"{seq}", file=sort_pipe.stdin)
                    else:
                        yield from result
            # Yield now if sort
            if self.sort:
                # Close pipe and yield result
                sort_pipe.stdin.close()
                yield from (line.strip() for line in sort_pipe.stdout)

    def _parallel_job(self, seq):
        """ Convert a single seq and return a list

        Parameters
        ----------
        seq : str
            A single input string to convert into kmers

        Returns
        -------
        [str]
            A list of kmer strings

        """
        sequences = (seq,)
        for f in self.parsers:
            sequences = f(sequences)
        return list(sequences)

    def __iter__(self):
        """ Return iterator of kmers from sequences provided at creation """
        return iter(self.__call__(self.sequences))

    def _parsed_kmers(self, sequences):
        """ Utility function to extract sequences from a file / iterable

        Parameters
        ----------
        sequences : [str] or str
            The iterable of sequences or a fasta filename

        Yields
        ------
        str
            An iterable of kmers

        """
        # Check if sequences is a file or an iterable
        if isinstance(sequences, str):
            # Filename
            sequences = self._read_file(sequences)

        # Check if fasta format and parse stream
        is_fasta, sequence = self._detect_FASTA(sequences)
        if is_fasta:
            sequences = self._parse_FASTA(sequences)
        else:
            sequences = self._parse_seqs(sequences)

        yield from sequences

    def _read_file(self, filename):
        """ Return iterator to file lines

        Parameters
        ----------
        filename : str
            The input Fasta file to read

        Yields
        ------
        str
            The lines of the input file

        """
        for line in fileinput.input(filename,
                                    openhook=fileinput.hook_compressed):
            # Try to decode
            try:
                tmp = line.decode()
                yield tmp
            except (UnicodeDecodeError, AttributeError):
                yield line

    def _detect_RNA(self, sequences):
        """ Checks if sequences is RNA or DNA, sets value and re-yields

        Parameters
        ----------
        sequences : Iterable of str
            The iterable of sequences

        Returns
        -------
        tuple of bool, iterable of str
            bool is True if sequences are RNA. Second return value is the
            Unmodified iterable of input sequences.

        """
        found = []
        is_rna = None
        for seq in sequences:
            # Add seq to queue
            found.append(seq)
            if 'T' in seq or 't' in seq:
                is_rna = False
                break
            if 'U' in seq or 'u' in seq:
                is_rna = True
                break
        # return tuple
        return (is_rna, itertools.chain(found, sequences))

    def _detect_FASTA(self, sequences):
        """ Checks if sequences is RNA or DNA, sets value and re-yields

        Parameters
        ----------
        sequences : generator of str
            The stream of lines to parse

        Returns
        ------
        tuple of bool, iterable of str
            bool is True if the sequence is from a Fasta file, else False
            The iterable is the unmodified sequence stream

        """
        lines = []
        is_fasta = None
        for seq in sequences:
            # Add seq to lines
            lines.append(seq)
            if '>' in seq:
                is_fasta = True
                break
            else:
                is_fasta = False
                break
        # return tuple
        return (is_fasta, itertools.chain(lines, sequences))

    def _parse_seqs(self, lines):
        """ Parses lines to remove spaces and newlines

        Parameters
        ----------
        lines : generator of str
            The stream of lines to parse

        Yields
        ------
        str
            The parsed stream of lines

        """
        for line in lines:
            yield line.strip()

    def _parse_FASTA(self, lines):
        """ Parse fasta file and yield sequences

        Parameters
        ----------
        lines : generator of str
            The stream of lines to convert into sequences

        Yields
        ------
        str
            A single sequence for each fasta entry

        """

        sequence = ""
        for line in lines:
            # Remove any newlines and spaces
            line = line.strip()
            if line.startswith('>'):
                if len(sequence):
                    yield sequence
                sequence = ""
            else:
                # Add line to sequence
                sequence += line
        if len(sequence):
            yield sequence

    def _to_DNA(self, sequences):
        """ Converts sequences to DNA

        Parameters
        ----------
        sequences : generator of str
            The stream of sequences to parse

        Yields
        ------
        str
            The parsed stream of sequences

        """
        return (seq.replace('U', 'T').replace('u', 't') for seq in sequences)

    def _to_RNA(self, sequences):
        """ Converts sequences to RNA

        Parameters
        ----------
        sequences : generator of str
            The stream of sequences to parse

        Yields
        ------
        str
            The parsed stream of sequences

        """
        return (seq.replace('T', 'U').replace('t', 'u') for seq in sequences)

    def _kmers(self, sequences):
        """ Convert sequecnes to kmers

        Parameters
        ----------
        sequences : generator of str
            The stream of sequences to parse

        Yields
        ------
        str
            The parsed stream of sequences

        """
        # Dummy funtion
        def _extract(seq):
            # Iterate through every length
            for k in self.kmers:
                # Iterate through each index and yield
                for i in range(len(seq) - k + 1):
                    tmp = [seq[i + j] for j in range(k)]
                    yield ''.join(tmp)

        # Extract every kmer from every sequence and yield
        for seq in sequences:
            yield from _extract(seq)

    def _get_complement(self, seq):
        """ Returns complement to a single sequence

        Parameters
        ----------
        seq : str
            sequence to get reverse complement of

        Returns
        -------
        str
            The reverse complement

        """
        new_seq = ''.join([COMP_MAP[b] for b in reversed(seq)])
        return new_seq

    def _complements(self, sequences):
        """ Add complements to stream

        Parameters
        ----------
        sequences : generator of str
            The stream of sequences to parse

        Yields
        ------
        str
            The parsed stream of sequences

        """
        for seq in sequences:
            yield seq
            yield self._get_complement(seq)

    def _canonicals(self, sequences):
        """ Only add canonical kmer to stream

        Parameters
        ----------
        sequences : generator of str
            The stream of sequences to parse

        Yields
        ------
        str
            The parsed stream of sequences

        """
        for seq in sequences:
            yield min(seq, self._get_complement(seq))

    def _allow(self, sequences):
        """ Return allowed sequences

        Parameters
        ----------
        sequences : generator of str
            The stream of sequences to parse

        Yields
        ------
        str
            The parsed stream of sequences

        """
        # Dummy function for filter
        def _isallowed(seq):
            return set(seq).issubset(self.allow)
        return filter(_isallowed, sequences)

    def _disallow(self, sequences):
        """ Omit disallowed sequences

        Parameters
        ----------
        sequences : generator of str
            The stream of sequences to parse

        Yields
        ------
        str
            The parsed stream of sequences

        """
        # Dummy function for filter
        def _isallowed(seq):
            return set(seq).isdisjoint(self.disallow)
        return filter(_isallowed, sequences)

    def _omitsoft(self, sequences):
        """ Omit sequences containing softmasked nucleotides

        Parameters
        ----------
        sequences : generator of str
            The stream of sequences to parse

        Yields
        ------
        str
            The parsed stream of sequences


        """
        return filter(lambda x: x.isupper(), sequences)

    def _mapsoft(self, sequences):
        """ Map sequences containing softmasking to uppercase

        Parameters
        ----------
        sequences : generator of str
            The stream of sequences to parse

        Yields
        ------
        str
            The parsed stream of sequences

        """
        # Map soft masked sequences to uppercase
        return (seq.upper() for seq in sequences)

    def _expandiupac(self, sequences):
        """ Expand iupac characters in sequences

        Parameters
        ----------
        sequences : generator of str
            The stream of sequences to parse

        Yields
        ------
        str
            The parsed stream of sequences

        """
        # Iterate through each sequence and call _expand
        for seq in sequences:
            # Find all IUPAC bases
            iupac_bases = []
            iupac_options = []
            for i, b in enumerate(seq):
                if b in IUPAC_BASE:
                    iupac_bases.append(i)
                    iupac_options.append(IUPAC_BASE[b])
            # Check IUPAC bases found:
            if len(iupac_bases):
                # Iterate through every combination of options
                seq = list(seq)
                for combo in itertools.product(*iupac_options):
                    # Place bases
                    for i, b in zip(iupac_bases, combo):
                        seq[i] = b
                    # yield
                    yield ''.join(seq)
            else:
                # No iupac so just yield
                yield seq

    def _split(self, sequences):
        """ Split sequence into ',' seperated columns

        Parameters
        ----------
        sequences : generator of str
            The stream of sequences to split

        Yields
        ------
        str
            The parsed stream of sequences

        """
        # Iterate through every sequence
        for seq in sequences:
            # Split seq into chunks
            pos_parts = []
            neg_parts = []
            for size in self.split:
                if size >= 0:
                    pos_parts.append(seq[:size])
                    seq = seq[size:]
                else:
                    neg_parts.append(seq[size:])
                    seq = seq[:size]
            # Merge parts and yield
            yield ','.join(pos_parts + [seq] + neg_parts)


def parseArgs(sys_args):
    """ Parse command line args """
    parser = argparse.ArgumentParser(
            description=("Read and parse kmers from fasta or kmer stream\n"
                         "Compatible with gz, bz2, and stdin."),
            prog="kstream",
            formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
            "file",
            nargs="?",
            type=str,
            help="Fasta file to read. .gz, .bz2, default stdin",
            default="-")
    parser.add_argument(
            "-k",
            "--kmers",
            type=int,
            nargs="+",
            help="Convert sequences into kmers of given length(s).")
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
            "--canonicals",
            action="store_true",
            help="Print canonical sequences (alphabetically first)")
    group.add_argument(
            "--complements",
            action="store_true",
            help="Add reverse complement to stream")
    parser.add_argument(
            "--disallow",
            type=str,
            help="Omit sequences containing dissallowed nucleotides")
    parser.add_argument(
            "--allow",
            type=str,
            help="Only accept sequences containing allowed nucleotides")
    parser.add_argument(
            "--expand-iupac",
            action="store_true",
            help="Expand IUPAC nucleotide codes (including N's)")
    parser.add_argument(
            "--omit-softmask",
            action="store_true",
            help="Omit sequences containing soft masking")
    parser.add_argument(
            "--map-softmask",
            action="store_true",
            help="Unmask sequences containing soft masking")
    parser.add_argument(
            "--split",
            nargs="+",
            type=int,
            help="Split kmers into columns and delimit by ','")
    parser.add_argument(
            "-p",
            "--parallel",
            type=int,
            default=1,
            help="Number of processors to use. Default 1")
    parser.add_argument(
            "-s",
            "--sort",
            action="store_true",
            help="Sort resulting kmers")
    parser.add_argument(
            "--sort-np",
            type=int,
            default=1,
            help="Number of processores to use for sorting")
    parser.add_argument(
            "--sort-mem",
            type=str,
            help="Amount of memory to use, see linux sort mem usage")
    parser.add_argument(
            "--sort-cols",
            nargs="+",
            type=int,
            help="Sort based on these columns, 0-based indexing")
    parser.add_argument(
            "--output",
            help="Write output to file as opposed to terminal")
    parser.add_argument(
            '--version',
            action='version',
            version='%(prog)s 1.0')

    # Parse args and return
    return parser.parse_args(sys_args)


def main():
    # Get command line args
    args = parseArgs(sys.argv[1:])

    # Create kstream instance
    streamer = kstream(kmers=args.kmers,
                       complements=args.complements,
                       canonicals=args.canonicals,
                       allow=args.allow,
                       disallow=args.disallow,
                       omitsoft=args.omit_softmask,
                       mapsoft=args.map_softmask,
                       expandiupac=args.expand_iupac,
                       split=args.split,
                       parallel=args.parallel,
                       sort=args.sort,
                       sortnp=args.sort_np,
                       sortmem=args.sort_mem,
                       sortcols=args.sort_cols)

    # Pass args.file into kstream and print or write result
    if args.output is not None:
        with open(args.output, "w") as fout:
            for seq in streamer(args.file):
                print(seq, file=fout)
    else:
        for seq in streamer(args.file):
            print(seq)


if __name__ == "__main__":
    main()
